"""
HTTP client implementation for Athena API.

This module provides HTTP clients for making requests to the Athena API,
with features like retry, backoff, and timeout handling.
"""

import json
import logging
import random
import time
from typing import Any, Dict, Optional, Tuple, TypeVar, Union

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .auth import build_headers
from .exceptions import (
    AthenaError,
    AuthenticationError,
    ClientError,
    NetworkError,
    RateLimitError,
    ServerError,
    TimeoutError,
    ValidationError,
)
from .settings import get_settings

# Type variable for generic response
T = TypeVar("T")

logger = logging.getLogger(__name__)


class HttpClient:
    """
    Synchronous HTTP client for making requests to the Athena API.

    Features:
    - Automatic retry with exponential backoff
    - Custom timeout handling
    - Authentication header generation
    - Error handling and mapping to typed exceptions
    - Robust logging and debugging
    - Custom User-Agent and headers
    """

    def __init__(
        self,
        base_url: Optional[str] = None,
        token: Optional[str] = None,
        client_id: Optional[str] = None,
        private_key: Optional[str] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        backoff_factor: Optional[float] = None,
        enable_throttling: bool = True,
        throttle_delay_range: Tuple[float, float] = (0.1, 0.3),
    ) -> None:
        """
        Initialize the HTTP client with configuration.

        Args:
            base_url: Base URL for the Athena API
            token: Bearer token for authentication
            client_id: Client ID for HMAC authentication
            private_key: Private key for HMAC signing
            timeout: HTTP timeout in seconds
            max_retries: Maximum number of retry attempts
            backoff_factor: Exponential backoff factor for retries
            enable_throttling: Whether to throttle requests to be respectful
                to the server
            throttle_delay_range: Range of delays for throttling (min, max)
                in seconds
        """
        settings = get_settings()

        # Use provided values or fall back to settings
        self.base_url = base_url or settings.ATHENA_BASE_URL

        # Set up token and HMAC if provided
        if token is not None:
            settings.ATHENA_TOKEN = token
        if client_id is not None:
            settings.ATHENA_CLIENT_ID = client_id
        if private_key is not None:
            settings.ATHENA_PRIVATE_KEY = private_key

        self.timeout = timeout or settings.ATHENA_TIMEOUT_SECONDS
        self.max_retries = max_retries or settings.ATHENA_MAX_RETRIES
        self.backoff_factor = backoff_factor or settings.ATHENA_BACKOFF_FACTOR
        self.enable_throttling = enable_throttling
        self.throttle_delay_range = throttle_delay_range

        # Create session with retry configuration
        self.session = self._create_session()

        # Set up default headers
        self._setup_default_headers()

        logger.debug("HttpClient initialized with base URL: %s", self.base_url)

    def _create_session(self) -> requests.Session:
        """
        Create and configure a requests Session with enhanced retry logic.

        Returns:
            Configured requests.Session object
        """
        session = requests.Session()

        # Enhanced retry strategy for better handling of rate limiting
        # and server overload
        retry_strategy = Retry(
            total=self.max_retries,
            backoff_factor=self.backoff_factor,
            # Retry on rate limiting (429), server errors (5xx), and temporary failures
            status_forcelist=[429, 500, 502, 503, 504, 520, 521, 522, 523, 524],
            # Also retry on connection errors and timeouts
            allowed_methods=["HEAD", "GET", "OPTIONS", "POST"],
            # Respect Retry-After headers from server
            respect_retry_after_header=True,
            # Exponential backoff with jitter to prevent thundering herd
            raise_on_status=False,  # Let us handle status codes ourselves
        )

        adapter = HTTPAdapter(max_retries=retry_strategy)
        session.mount("http://", adapter)
        session.mount("https://", adapter)

        return session

    def _setup_default_headers(self) -> None:
        """Set up default headers for all requests."""
        # Set default headers similar to the working client
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": "AthenaOHDSIAPIClient/1.0",
        }

        # Update session headers
        self.session.headers.update(default_headers)

        logger.debug("Default headers set: %s", default_headers)

    def _throttle_request(self) -> None:
        """
        Implement request throttling to prevent overwhelming the server.

        This adds a small delay between requests to be respectful to the API.
        """
        if not self.enable_throttling:
            return

        # Add a small random delay between requests
        # This prevents overwhelming the server with rapid requests
        delay = random.uniform(  # nosec B311
            self.throttle_delay_range[0], self.throttle_delay_range[1]
        )
        time.sleep(delay)

        logger.debug(f"Request throttled for {delay:.3f} seconds")

    def _handle_rate_limit(self, response: requests.Response) -> None:
        """
        Handle rate limiting with intelligent backoff.

        Args:
            response: HTTP response that indicates rate limiting
        """
        # Get retry-after header if available
        retry_after = response.headers.get("Retry-After")

        if retry_after:
            try:
                wait_time = int(retry_after)
                logger.info(
                    f"Rate limited. Waiting {wait_time} seconds as requested by server."
                )
                time.sleep(wait_time)
            except ValueError:
                # If Retry-After is not a number, use exponential backoff
                wait_time = 60  # Default to 1 minute
                logger.info(f"Rate limited. Waiting {wait_time} seconds (default).")
                time.sleep(wait_time)
        else:
            # No Retry-After header, use exponential backoff
            wait_time = 60  # Default to 1 minute
            logger.info(
                f"Rate limited. Waiting {wait_time} seconds (exponential backoff)."
            )
            time.sleep(wait_time)

    def _build_url(self, path: str) -> str:
        """
        Build full URL by joining base URL and path.

        Args:
            path: API path

        Returns:
            Full URL
        """
        # Handle paths that start with / to ensure they're appended correctly
        if path.startswith("/"):
            # Remove the leading / and join with base_url
            path = path[1:]

        # Ensure base_url doesn't end with / and path doesn't start with /
        if self.base_url.endswith("/"):
            base = self.base_url[:-1]
        else:
            base = self.base_url

        if path.startswith("/"):
            path = path[1:]

        full_url = f"{base}/{path}"

        logger.debug(
            f"Building URL: base_url='{self.base_url}', path='{path}' "
            f"-> full_url='{full_url}'"
        )
        return full_url

    def _normalize_params(
        self, params: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, str]]:
        """
        Normalize parameters to ensure all values are strings.

        Args:
            params: Query parameters

        Returns:
            Normalized parameters with string values
        """
        if params is None:
            return None

        normalized = {}
        for key, value in params.items():
            if value is not None:
                normalized[key] = str(value)
        return normalized

    def _handle_response(self, response: requests.Response, url: str) -> Dict[str, Any]:
        """
        Handle API response and raise appropriate exceptions.

        Args:
            response: HTTP response from requests
            url: Full URL that was requested

        Returns:
            Parsed JSON response

        Raises:
            ClientError: For 4xx status codes
            ServerError: For 5xx status codes
            NetworkError: For connection errors
            APIError: For API-specific error responses
        """
        # Log raw response for debugging
        raw_response_text = response.text
        logger.debug(f"Raw response text from {url}: {raw_response_text[:1000]}...")

        try:
            response.raise_for_status()

            # Attempt to parse JSON after logging raw text
            data = response.json()
            logger.debug("Successfully parsed JSON from %s", url)
            return data

        except requests.exceptions.HTTPError as e:
            status = e.response.status_code if e.response else "N/A"
            text = e.response.text if e.response else "No response content"

            # Try to parse the error response as JSON to extract API error details
            api_message = None

            try:
                if e.response and e.response.text:
                    error_data = e.response.json()
                    api_message = error_data.get("errorMessage")
            except (ValueError, AttributeError):
                pass

            # Create a more descriptive error message
            if api_message:
                msg = f"API Error {status}: {api_message}"
            else:
                msg = f"HTTP error {status} when accessing {url}"
                if text and text != "No response content":
                    msg += f": {text[:200]}"

            logger.error(msg)
            logger.exception(e)

            # Raise specific exception types based on status code
            if status == 401:
                raise AuthenticationError(
                    f"Authentication failed: {msg}",
                    status_code=response.status_code,
                    response=response.text,
                    url=url,
                ) from e
            elif status == 403:
                raise ClientError(
                    f"Access forbidden: {msg}",
                    status_code=response.status_code,
                    response=response.text,
                    url=url,
                ) from e
            elif status == 404:
                raise ClientError(
                    f"Resource not found: {msg}",
                    status_code=response.status_code,
                    response=response.text,
                    url=url,
                ) from e
            elif status == 429:
                raise RateLimitError(
                    f"Rate limit exceeded: {msg}",
                    status_code=response.status_code,
                    response=response.text,
                    url=url,
                ) from e
            elif 400 <= response.status_code < 500:
                raise ClientError(
                    f"Client error: {msg}",
                    status_code=response.status_code,
                    response=response.text,
                    url=url,
                ) from e
            elif response.status_code >= 500:
                raise ServerError(
                    f"Server error: {msg}",
                    status_code=response.status_code,
                    response=response.text,
                    url=url,
                ) from e
            else:
                raise

        except json.JSONDecodeError as e:
            msg = (
                f"Invalid JSON response from {url}: {e}. "
                f"Raw text was: {raw_response_text[:1000]}..."
            )
            logger.error(msg)
            logger.exception(e)
            raise ValidationError(
                f"Invalid JSON response: {e}", validation_details=str(e)
            ) from e

        except Exception as e:
            msg = f"An unexpected error occurred when accessing {url}: {e}"
            logger.error(msg)
            logger.exception(e)
            raise AthenaError(msg, error_code="UNEXPECTED_ERROR") from e

    def request(
        self,
        method: str,
        path: str,
        data: Any = None,
        params: Optional[Dict[str, Any]] = None,
        raw_response: bool = False,
    ) -> Union[Dict[str, Any], requests.Response]:
        """
        Make an HTTP request to the Athena API with enhanced retry and throttling.

        Args:
            method: HTTP method (GET, POST, etc.)
            path: API endpoint path
            params: Query parameters
            data: Request body data
            raw_response: Whether to return the raw response object

        Returns:
            Parsed JSON response or raw Response object

        Raises:
            ClientError: For 4xx status codes
            ServerError: For 5xx status codes
            NetworkError: For connection errors
            RateLimitError: For rate limiting issues
        """
        url = self._build_url(path)
        body_bytes = b""

        # Convert data to JSON bytes if provided
        if data is not None:
            body_bytes = json.dumps(data).encode("utf-8")

        # Build authentication headers
        auth_headers = build_headers(method, url, body_bytes)

        # Merge with session headers
        headers = dict(self.session.headers)
        headers.update(auth_headers)

        # Normalize parameters to strings
        normalized_params = self._normalize_params(params)

        # Generate a correlation ID for logging
        correlation_id = f"req-{id(self)}-{id(path)}"
        logger.debug(
            f"[{correlation_id}] {method} {url} with params: {normalized_params}"
        )

        # Throttle request to be respectful to the server
        self._throttle_request()

        try:
            response = self.session.request(
                method=method,
                url=url,
                params=normalized_params,
                data=body_bytes if data is not None else None,
                headers=headers,
                timeout=self.timeout,
            )

            logger.debug(f"[{correlation_id}] {response.status_code} {response.reason}")

            # Handle rate limiting specifically
            if response.status_code == 429:
                logger.warning(
                    f"Rate limited by server. Status: {response.status_code}"
                )
                self._handle_rate_limit(response)
                # After waiting, we could retry the request, but for now we'll
                # raise the error
                # The retry mechanism in the client will handle this

            if raw_response:
                return response

            return self._handle_response(response, url)

        except requests.exceptions.Timeout as e:
            msg = (
                f"Timeout when accessing {url}. "
                f"Verify network connectivity and the API's responsiveness."
            )
            logger.error(msg)
            logger.exception(e)
            raise TimeoutError(msg, url=url, timeout=self.timeout) from e

        except requests.exceptions.ConnectionError as e:
            msg = (
                f"Connection error when accessing {url}. "
                f"Check your network connection and endpoint URL."
            )
            logger.error(msg)
            logger.exception(e)
            raise NetworkError(msg, url=url) from e

        except Exception as e:
            msg = f"An unexpected error occurred when accessing {url}: {e}"
            logger.error(msg)
            logger.exception(e)
            raise NetworkError(msg, url=url) from e

    def get(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        raw_response: bool = False,
    ) -> Union[Dict[str, Any], requests.Response]:
        """
        Make a GET request to the Athena API.

        Args:
            path: API endpoint path
            params: Query parameters
            raw_response: Whether to return the raw response object

        Returns:
            Parsed JSON response or raw Response object
        """
        return self.request("GET", path, params=params, raw_response=raw_response)

    def post(
        self,
        path: str,
        data: Dict[str, Any],
        params: Optional[Dict[str, Any]] = None,
        raw_response: bool = False,
    ) -> Union[Dict[str, Any], requests.Response]:
        """
        Make a POST request to the Athena API.

        Args:
            path: API endpoint path
            data: Request body data
            params: Query parameters
            raw_response: Whether to return the raw response object

        Returns:
            Parsed JSON response or raw Response object
        """
        return self.request(
            "POST", path, data=data, params=params, raw_response=raw_response
        )

    def close(self) -> None:
        """Closes the underlying HTTP session."""
        logger.info("Closing HTTP session.")
        self.session.close()

    def __enter__(self) -> "HttpClient":
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.close()



import logging
from typing import Any, Dict, List, Optional

from .concept_explorer import ConceptExplorer
from .db.base import DatabaseConnector

logger = logging.getLogger(__name__)


class ConceptSetGenerator:
    """Orchestrate concept set generation from a search query."""

    def __init__(self, explorer: ConceptExplorer, db: DatabaseConnector) -> None:
        self.explorer = explorer
        self.db = db

    async def create_from_query(
        self,
        query: str,
        strategy: str = "fallback",
        include_descendants: bool = True,
        confidence_threshold: float = 0.7,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Generate a concept set using a tiered standard-first strategy.
        Includes robust error handling.
        """
        from .exceptions import AthenaError

        try:
            logger.info(
                "Attempting to find and validate standard concepts for query: '%s'",
                query,
            )
            mappings = await self.explorer.map_to_standard_concepts(
                query, confidence_threshold=confidence_threshold, **kwargs
            )

            from .models import ConceptType

            standard_concepts = [
                m["concept"]
                for m in mappings
                if m["concept"].standardConcept == ConceptType.STANDARD
            ]

            if standard_concepts:
                candidate_ids = [c.id for c in standard_concepts]
                validated_ids = self.db.validate_concepts(candidate_ids)

                if validated_ids:
                    warnings: List[str] = []
                    final_ids = set(validated_ids)
                    if include_descendants:
                        desc_ids = self.db.get_descendants(validated_ids)
                        if not desc_ids:
                            warnings.append(
                                (
                                    f"Standard concepts {validated_ids} found but "
                                    "have no descendants in the local vocabulary."
                                )
                            )
                        else:
                            final_ids.update(desc_ids)

                    return self._build_success_response(
                        query=query,
                        strategy_used="Tier 1: Direct Standard Concept",
                        concept_ids=list(final_ids),
                        seed_concepts=validated_ids,
                        warnings=warnings,
                    )

            if strategy == "strict":
                return self._build_failure_response(
                    query,
                    (
                        "No standard concepts from the API could be validated in "
                        "'strict' mode."
                    ),
                )

            non_standard_concepts = [
                m["concept"]
                for m in mappings
                if m["concept"].standardConcept != ConceptType.STANDARD
            ]

            if non_standard_concepts:
                local_mappings = self.db.get_standard_mapping(
                    [c.id for c in non_standard_concepts]
                )
                if local_mappings:
                    mapped_standard_id = list(local_mappings.values())[0]
                    validated_ids = self.db.validate_concepts([mapped_standard_id])
                    if validated_ids:
                        final_ids = set(validated_ids)
                        if include_descendants:
                            final_ids.update(self.db.get_descendants(validated_ids))
                        original_non_standard = list(local_mappings.keys())[0]
                        warning = (
                            "Initial standard concepts not found locally. "
                            "Recovered by mapping non-standard concept "
                            f"{original_non_standard} "
                            f"to standard concept {mapped_standard_id}."
                        )
                        return self._build_success_response(
                            query=query,
                            strategy_used="Tier 3: Recovery via Local Mapping",
                            concept_ids=list(final_ids),
                            seed_concepts=validated_ids,
                            warnings=[warning],
                        )

            return self._build_failure_response(
                query,
                (
                    "No standard concepts from the API could be validated against "
                    "the local database, and no recovery paths were found."
                ),
            )
        except AthenaError as err:
            logger.error(f"Athena client error during concept set generation: {err}")
            # Return a structured failure response with troubleshooting info
            return {
                "name": f"Failed Concept Set for '{query}'",
                "concept_ids": [],
                "metadata": {
                    "status": "FAILURE",
                    "source_query": query,
                    "reason": str(err),
                    "suggestions": [
                        "Check your network connection and credentials.",
                        "Try again in a few moments.",
                        getattr(err, "troubleshooting", None)
                        or "See logs for more details.",
                    ],
                },
            }

    def _build_success_response(
        self,
        query: str,
        strategy_used: str,
        concept_ids: List[int],
        seed_concepts: List[int],
        warnings: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        return {
            "name": f"Concept Set for '{query}'",
            "concept_ids": concept_ids,
            "metadata": {
                "status": "SUCCESS",
                "strategy_used": strategy_used,
                "source_query": query,
                "seed_concepts": seed_concepts,
                "warnings": warnings or [],
            },
        }

    def _build_failure_response(self, query: str, reason: str) -> Dict[str, Any]:
        return {
            "name": f"Failed Concept Set for '{query}'",
            "concept_ids": [],
            "metadata": {
                "status": "FAILURE",
                "source_query": query,
                "reason": reason,
                "suggestions": [
                    "Try a different search term.",
                    "Lower the `confidence_threshold`.",
                    "Check if your local vocabulary version is up to date.",
                ],
            },
        }



"""
Advanced concept exploration and mapping utilities.

This module provides tools to help find standard concepts that might not appear
directly in search results, including:
- Concept relationship exploration
- Synonym-based concept discovery
- Vocabulary mapping and cross-references
- Standard concept identification
- Concept hierarchy exploration
"""

import asyncio
import logging
from collections import deque
from dataclasses import dataclass, field
from typing import Any, Deque, Dict, List, Optional, Set, Union

from .models import Concept, ConceptDetails

logger = logging.getLogger(__name__)


@dataclass
class ExplorationState:
    """
    Internal state management for concept exploration runs.

    This dataclass manages the state of a single exploration run to avoid
    passing multiple state variables through method calls.
    """

    queue: Deque[Any] = field(
        default_factory=deque
    )  # BFS queue: (concept, depth, path)
    visited_ids: Set[int] = field(default_factory=set)  # Set of processed concept IDs
    cache: Dict[int, ConceptDetails] = field(
        default_factory=dict
    )  # In-memory cache for concept details
    results: Dict[str, Any] = field(
        default_factory=lambda: {
            "direct_matches": [],
            "synonym_matches": [],
            "relationship_matches": [],
            "cross_references": [],
            "exploration_paths": [],
        }
    )  # Final results structure
    api_call_count: int = 0  # Count of API calls made
    start_time: float = 0  # Start time of exploration
    max_total_concepts: int = 0  # Maximum total concepts to discover
    max_api_calls: int = 0  # Maximum API calls to make
    max_time_seconds: int = 0  # Maximum time to run exploration

    def __post_init__(self) -> None:
        """Initialize default values for dataclass fields."""
        pass  # All defaults handled by field()


class ConceptExplorer:
    """
    Advanced concept exploration and mapping utilities.

    This class provides methods to help discover standard concepts that might
    not appear directly in search results by exploring relationships, synonyms,
    and cross-references using an optimized BFS algorithm with concurrent processing.
    """

    def __init__(self, client: Any) -> None:
        """Initialize the concept explorer with an Athena client.

        Args:
            client: Athena client instance (can be sync or async)
        """
        self.client = client

    async def find_standard_concepts(
        self,
        query: str,
        max_exploration_depth: int = 2,
        initial_seed_limit: int = 10,  # New parameter per FSD
        include_synonyms: bool = True,
        include_relationships: bool = True,
        vocabulary_priority: Optional[List[str]] = None,
        max_total_concepts: int = 100,  # Limit total concepts found
        max_api_calls: int = 50,  # Limit total API calls
        max_time_seconds: int = 30,  # Limit total execution time
    ) -> Dict[str, Any]:
        """
        Find standard concepts related to a query through exploration.

        This method implements a BFS-based exploration algorithm with concurrent
        batch processing for optimal performance.

        Args:
            query: The search query
            max_exploration_depth: Maximum depth to explore relationships
            initial_seed_limit: Maximum number of top search results to use as seeds
            include_synonyms: Whether to explore synonyms
            include_relationships: Whether to explore relationships
            vocabulary_priority: Preferred vocabularies (e.g., ['SNOMED', 'RxNorm'])
            max_total_concepts: Maximum total concepts to discover
            max_api_calls: Maximum API calls to make
            max_time_seconds: Maximum time to run exploration

        Returns:
            Dictionary containing standard concepts found through exploration
        """
        # Validate parameters
        if max_exploration_depth < 0:
            raise ValueError("max_exploration_depth must be non-negative")
        if initial_seed_limit < 1:
            raise ValueError("initial_seed_limit must be at least 1")
        if max_total_concepts < 1:
            raise ValueError("max_total_concepts must be at least 1")
        if max_api_calls < 1:
            raise ValueError("max_api_calls must be at least 1")
        if max_time_seconds < 1:
            raise ValueError("max_time_seconds must be at least 1")

        # Initialize exploration state with limits
        state = ExplorationState()
        state.api_call_count = 0
        state.start_time = asyncio.get_event_loop().time()
        state.max_total_concepts = max_total_concepts
        state.max_api_calls = max_api_calls
        state.max_time_seconds = max_time_seconds

        # Step 1: Initial search and seeding
        logger.info(f"Performing direct search for: {query}")
        direct_results = await self.client.search(query, size=50)
        state.api_call_count += 1
        all_direct_matches = direct_results.all()
        state.results["direct_matches"] = all_direct_matches

        # Take top initial_seed_limit results as seeds
        seed_concepts = all_direct_matches[:initial_seed_limit]
        logger.info(f"Using {len(seed_concepts)} concepts as exploration seeds")

        # Add seed concepts to queue and mark as visited
        for concept in seed_concepts:
            state.queue.append((concept, 0, [concept.id]))
            state.visited_ids.add(concept.id)

        # Step 2: Main BFS traversal loop
        await self._bfs_exploration_loop(
            state,
            max_exploration_depth,
            include_synonyms,
            include_relationships,
            vocabulary_priority,
        )

        # Step 3: Find cross-references (with limits)
        if state.api_call_count < state.max_api_calls:
            logger.info("Finding cross-references")
            cross_refs = await self._find_cross_references(
                all_direct_matches, vocabulary_priority, state
            )
            state.results["cross_references"] = cross_refs

        # Add exploration statistics
        elapsed_time = asyncio.get_event_loop().time() - state.start_time
        state.results["exploration_stats"] = {
            "total_concepts_found": len(state.visited_ids),
            "api_calls_made": state.api_call_count,
            "time_elapsed_seconds": elapsed_time,
            "limits_reached": {
                "max_concepts": len(state.visited_ids) >= max_total_concepts,
                "max_api_calls": state.api_call_count >= max_api_calls,
                "max_time": elapsed_time >= max_time_seconds,
            },
        }

        return state.results

    async def _bfs_exploration_loop(
        self,
        state: ExplorationState,
        max_exploration_depth: int,
        include_synonyms: bool,
        include_relationships: bool,
        vocabulary_priority: Optional[List[str]] = None,
    ) -> None:
        """
        Main BFS exploration loop with concurrent batch processing.

        Args:
            state: Exploration state object
            max_exploration_depth: Maximum exploration depth
            include_synonyms: Whether to explore synonyms
            include_relationships: Whether to explore relationships
            vocabulary_priority: Preferred vocabularies
        """
        while state.queue:
            # Check limits before processing
            current_time = asyncio.get_event_loop().time()
            elapsed_time = current_time - state.start_time

            # Check if we've hit any limits
            if (
                len(state.visited_ids) >= state.max_total_concepts
                or state.api_call_count >= state.max_api_calls
                or elapsed_time >= state.max_time_seconds
            ):
                logger.info("Exploration stopped due to limits:")
                logger.info(
                    f"  - Concepts found: {len(state.visited_ids)}/"
                    f"{state.max_total_concepts}"
                )
                logger.info(
                    f"  - API calls made: {state.api_call_count}/{state.max_api_calls}"
                )
                logger.info(
                    f"  - Time elapsed: {elapsed_time:.1f}s/{state.max_time_seconds}s"
                )
                break

            # Process all nodes at current depth level
            level_size = len(state.queue)
            ids_to_fetch_details: Set[int] = set()

            logger.info(f"Processing {level_size} concepts at current depth level")

            # Process all nodes at current depth
            for _ in range(level_size):
                # Check limits again in case we've hit them during processing
                if (
                    len(state.visited_ids) >= state.max_total_concepts
                    or state.api_call_count >= state.max_api_calls
                    or asyncio.get_event_loop().time() - state.start_time
                    >= state.max_time_seconds
                ):
                    break

                concept, depth, path = state.queue.popleft()

                # Skip if we've reached max depth
                if depth >= max_exploration_depth:
                    continue

                # Discover neighbors (synonyms and relationships)
                await self._discover_neighbors(
                    concept,
                    depth,
                    path,
                    state,
                    ids_to_fetch_details,
                    include_synonyms,
                    include_relationships,
                )

            # Batch fetch details for all discovered concepts
            if ids_to_fetch_details and state.api_call_count < state.max_api_calls:
                await self._process_batch_results(
                    ids_to_fetch_details, state, depth + 1, vocabulary_priority
                )

    async def _discover_neighbors(
        self,
        concept: Concept,
        depth: int,
        path: List[int],
        state: ExplorationState,
        ids_to_fetch_details: Set[int],
        include_synonyms: bool,
        include_relationships: bool,
    ) -> None:
        """
        Discover neighboring concepts through synonyms and relationships.

        Args:
            concept: Current concept being explored
            depth: Current exploration depth
            path: Path taken to reach this concept
            state: Exploration state
            ids_to_fetch_details: Set to collect IDs for batch fetching
            include_synonyms: Whether to explore synonyms
            include_relationships: Whether to explore relationships
        """
        # Explore synonyms if enabled
        if include_synonyms:
            await self._discover_synonym_neighbors(
                concept, depth, path, state, ids_to_fetch_details
            )

        # Explore relationships if enabled
        if include_relationships:
            await self._discover_relationship_neighbors(
                concept, depth, path, state, ids_to_fetch_details
            )

    async def _discover_synonym_neighbors(
        self,
        concept: Concept,
        depth: int,
        path: List[int],
        state: ExplorationState,
        ids_to_fetch_details: Set[int],
    ) -> None:
        """Discover neighboring concepts through synonyms."""
        if concept.standardConcept == "Standard":
            # For standard concepts, search for synonyms
            try:
                # Check API call limit
                if state.api_call_count >= state.max_api_calls:
                    return

                synonym_results = await self.client.search(concept.name, size=10)
                state.api_call_count += 1

                for synonym_concept in synonym_results.all():
                    if (
                        synonym_concept.id not in state.visited_ids
                        and len(state.visited_ids) < state.max_total_concepts
                    ):
                        ids_to_fetch_details.add(synonym_concept.id)

            except Exception as e:
                logger.warning(
                    f"Could not generate suggestions for query '{concept.name}': {e}"
                )

    async def _discover_relationship_neighbors(
        self,
        concept: Concept,
        depth: int,
        path: List[int],
        state: ExplorationState,
        ids_to_fetch_details: Set[int],
    ) -> None:
        """Discover neighboring concepts through relationships."""
        try:
            # Check API call limit
            if state.api_call_count >= state.max_api_calls:
                return

            relationships = await self.client.relationships(
                concept.id, only_standard=True
            )
            state.api_call_count += 1

            for group in relationships.items:
                for rel in group.relationships:
                    if (
                        rel.targetConceptId not in state.visited_ids
                        and len(state.visited_ids) < state.max_total_concepts
                    ):
                        ids_to_fetch_details.add(rel.targetConceptId)

        except Exception as e:
            logger.warning(f"Could not get relationships for concept {concept.id}: {e}")

    async def _process_batch_results(
        self,
        ids_to_fetch_details: Set[int],
        state: ExplorationState,
        next_depth: int,
        vocabulary_priority: Optional[List[str]] = None,
    ) -> None:
        """Process batch results and add to queue."""
        if not ids_to_fetch_details:
            return

        # Check API call limit
        if state.api_call_count >= state.max_api_calls:
            return

        try:
            # Batch fetch concept details
            concept_details_list = await self._get_details_batch_async(
                ids_to_fetch_details
            )
            state.api_call_count += 1

            # Process each concept detail
            for concept_details in concept_details_list:
                if isinstance(concept_details, BaseException):
                    continue

                # Now we know it's a ConceptDetails object
                concept_details_obj = concept_details  # type: ConceptDetails

                # Convert to Concept object
                concept = Concept(
                    id=concept_details_obj.id,
                    name=concept_details_obj.name,
                    domain=concept_details_obj.domainId,
                    vocabulary=concept_details_obj.vocabularyId,
                    className=concept_details_obj.conceptClassId,
                    standardConcept=concept_details_obj.standardConcept,
                    code=concept_details_obj.conceptCode,
                    invalidReason=concept_details_obj.invalidReason,
                    score=None,  # No score available from concept details
                )

                # Add to appropriate result category
                if concept.standardConcept == "Standard":
                    if concept not in state.results["synonym_matches"]:
                        state.results["synonym_matches"].append(concept)
                else:
                    if concept not in state.results["relationship_matches"]:
                        state.results["relationship_matches"].append(concept)

                # Add to queue for further exploration if within limits
                if (
                    len(state.visited_ids) < state.max_total_concepts
                    and concept.id not in state.visited_ids
                ):
                    # Create a simple path for this concept
                    new_path = [concept.id]
                    state.queue.append((concept, next_depth, new_path))
                    state.visited_ids.add(concept.id)

        except Exception as e:
            logger.warning(f"Error processing batch results: {e}")

    async def _get_details_batch_async(
        self, concept_ids: Set[int]
    ) -> List[Union[ConceptDetails, BaseException]]:
        """
        Fetch concept details for multiple IDs concurrently.

        Args:
            concept_ids: Set of concept IDs to fetch details for

        Returns:
            List of ConceptDetails objects or Exception objects for failed requests
        """
        if not (
            hasattr(self.client, "get_concept_details")
            and asyncio.iscoroutinefunction(self.client.get_concept_details)
        ):
            raise NotImplementedError("Async client not available")

        # Create tasks for concurrent execution
        tasks = [
            self.client.get_concept_details(concept_id) for concept_id in concept_ids
        ]

        # Execute all tasks concurrently with exception handling
        results = await asyncio.gather(*tasks, return_exceptions=True)
        return results

    def _extract_standard_concepts(
        self, concepts: List[Concept], vocabulary_priority: Optional[List[str]] = None
    ) -> List[Concept]:
        """Extract standard concepts from a list of concepts."""
        standard_concepts = []

        for concept in concepts:
            if concept.standardConcept == "Standard":
                standard_concepts.append(concept)

        # Sort by vocabulary priority if specified
        if vocabulary_priority:

            def sort_key(concept: Concept) -> int:
                try:
                    return vocabulary_priority.index(concept.vocabulary)
                except ValueError:
                    return len(vocabulary_priority)

            standard_concepts.sort(key=sort_key)

        return standard_concepts

    async def _find_cross_references(
        self,
        concepts: List[Concept],
        vocabulary_priority: Optional[List[str]] = None,
        state: Optional[ExplorationState] = None,
    ) -> List[Dict[str, Any]]:
        """Find cross-references between concepts."""
        cross_refs = []

        for concept in concepts:
            if concept.standardConcept == "Standard":
                # Look for related concepts in other vocabularies
                try:
                    # Check API call limit
                    if state and state.api_call_count >= state.max_api_calls:
                        break

                    relationships = await self.client.relationships(
                        concept.id, only_standard=True
                    )
                    if state:
                        state.api_call_count += 1

                    for group in relationships.items:
                        for rel in group.relationships:
                            cross_ref = {
                                "source_concept": concept,
                                "target_concept_id": rel.targetConceptId,
                                "target_concept_name": rel.targetConceptName,
                                "relationship_type": rel.relationshipName,
                                "vocabulary": rel.targetVocabularyId,
                            }
                            cross_refs.append(cross_ref)

                except Exception as e:
                    logger.warning(
                        f"Could not get relationships for concept {concept.id}: {e}"
                    )

        return cross_refs

    async def map_to_standard_concepts(
        self,
        query: str,
        target_vocabularies: Optional[List[str]] = None,
        confidence_threshold: float = 0.5,
    ) -> List[Dict[str, Any]]:
        """
        Map a query to standard concepts with confidence scores.

        This method is now async to support the new async find_standard_concepts.

        Args:
            query: The search query
            target_vocabularies: Preferred target vocabularies
            confidence_threshold: Minimum confidence score for inclusion

        Returns:
            List of mapping results with confidence scores
        """
        # Get exploration results
        exploration_results = await self.find_standard_concepts(query)

        # Collect all standard concepts
        all_standard_concepts = []
        all_standard_concepts.extend(exploration_results["direct_matches"])
        all_standard_concepts.extend(exploration_results["synonym_matches"])
        all_standard_concepts.extend(exploration_results["relationship_matches"])

        # Filter by target vocabularies if specified
        if target_vocabularies:
            all_standard_concepts = [
                concept
                for concept in all_standard_concepts
                if concept.vocabulary in target_vocabularies
            ]

        # Calculate confidence scores and create mappings
        mappings = []
        for concept in all_standard_concepts:
            confidence = self._calculate_mapping_confidence(
                query, concept, exploration_results
            )

            if confidence >= confidence_threshold:
                mapping = {
                    "concept": concept,
                    "confidence": confidence,
                    "exploration_path": self._get_exploration_path(
                        concept, exploration_results
                    ),
                    "source_category": self._get_source_category(
                        concept, exploration_results
                    ),
                }
                mappings.append(mapping)

        # Sort by confidence score (descending)
        mappings.sort(key=lambda x: x["confidence"], reverse=True)

        return mappings

    def _calculate_mapping_confidence(
        self, query: str, concept: Concept, exploration_results: Dict[str, Any]
    ) -> float:
        """Calculate confidence score for a concept mapping."""
        confidence = 0.0

        # Base confidence from search score if available
        if concept.score is not None:
            confidence += concept.score * 0.4

        # Boost for direct matches
        if concept in exploration_results["direct_matches"]:
            confidence += 0.3

        # Boost for synonym matches
        if concept in exploration_results["synonym_matches"]:
            confidence += 0.2

        # Boost for relationship matches
        if concept in exploration_results["relationship_matches"]:
            confidence += 0.1

        # Vocabulary preference boost
        if concept.vocabulary in ["SNOMED", "RxNorm"]:
            confidence += 0.1

        return min(confidence, 1.0)

    def _get_exploration_path(
        self, concept: Concept, exploration_results: Dict[str, Any]
    ) -> str:
        """Get the exploration path that led to this concept."""
        if concept in exploration_results["direct_matches"]:
            return "direct_search"
        elif concept in exploration_results["synonym_matches"]:
            return "synonym_exploration"
        elif concept in exploration_results["relationship_matches"]:
            return "relationship_exploration"
        else:
            return "unknown"

    def _get_source_category(
        self, concept: Concept, exploration_results: Dict[str, Any]
    ) -> str:
        """Get the source category for a concept."""
        if concept in exploration_results["direct_matches"]:
            return "direct_match"
        elif concept in exploration_results["synonym_matches"]:
            return "synonym_match"
        elif concept in exploration_results["relationship_matches"]:
            return "relationship_match"
        else:
            return "unknown"

    async def suggest_alternative_queries(
        self, query: str, max_suggestions: int = 5
    ) -> List[str]:
        """
        Suggest alternative queries based on the original query.

        Args:
            query: The original query
            max_suggestions: Maximum number of suggestions to return

        Returns:
            List of alternative query suggestions
        """
        suggestions = []

        # Basic query variations
        base_suggestions = [
            query.lower(),
            query.title(),
            query.upper(),
            f"{query}*",  # Wildcard
            f"*{query}*",  # Contains
        ]

        # Add base suggestions
        suggestions.extend(base_suggestions)

        # Try to find related concepts for more suggestions
        try:
            results = await self.client.search(query, size=10)
            for concept in results.all():
                if concept.name.lower() != query.lower():
                    suggestions.append(concept.name)
        except Exception as e:
            logger.warning(f"Could not generate suggestions for query '{query}': {e}")

        # Remove duplicates and limit results
        unique_suggestions = list(dict.fromkeys(suggestions))
        return unique_suggestions[:max_suggestions]

    async def get_concept_hierarchy(
        self, concept_id: int, max_depth: int = 3
    ) -> Dict[str, Any]:
        """
        Get the concept hierarchy for a given concept.

        Args:
            concept_id: The concept ID
            max_depth: Maximum depth to explore

        Returns:
            Dictionary containing hierarchy information
        """
        hierarchy: Dict[str, Any] = {
            "root_concept": None,
            "parents": [],
            "children": [],
            "siblings": [],
            "depth": 0,
        }

        try:
            # Get root concept details
            root_details = await self.client.details(concept_id)
            hierarchy["root_concept"] = root_details

            # Initialize typed lists
            parents: List[ConceptDetails] = []
            children: List[ConceptDetails] = []

            # Get relationships
            relationships = await self.client.relationships(
                concept_id, only_standard=True
            )

            for group in relationships.items:
                if group.relationshipName.lower() in ["is a", "isa", "subclass of"]:
                    # Parent relationships
                    for rel in group.relationships:
                        try:
                            parent_details = await self.client.details(
                                rel.targetConceptId
                            )
                            parents.append(parent_details)
                        except Exception as e:
                            logger.warning(
                                f"Could not fetch parent concept "
                                f"{rel.targetConceptId}: {e}"
                            )

                elif group.relationshipName.lower() in ["has subclass", "has subtype"]:
                    # Child relationships
                    for rel in group.relationships:
                        try:
                            child_details = await self.client.details(
                                rel.targetConceptId
                            )
                            children.append(child_details)
                        except Exception as e:
                            logger.warning(
                                f"Could not fetch child concept "
                                f"{rel.targetConceptId}: {e}"
                            )

            # Update hierarchy with typed lists
            hierarchy["parents"] = parents
            hierarchy["children"] = children
            hierarchy["depth"] = len(parents)

        except Exception as e:
            logger.error(f"Could not get hierarchy for concept {concept_id}: {e}")

        return hierarchy


def create_concept_explorer(client: Any) -> ConceptExplorer:
    """
    Create a ConceptExplorer instance with the given client.

    Args:
        client: Athena client instance

    Returns:
        ConceptExplorer instance
    """
    return ConceptExplorer(client)



"""
Pydantic models for Athena API responses.

This module defines Pydantic models for the various responses from the Athena API.
"""

from enum import Enum
from typing import Any, ClassVar, Dict, List, Optional, Union, cast

import orjson
from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, Field, model_validator


def _json_dumps(value: Any, *, default: Any) -> str:
    """Serialize to JSON using orjson."""
    return orjson.dumps(value, default=default).decode()


class BaseModel(PydanticBaseModel):
    """Project-wide Pydantic base model using orjson."""

    model_config: ClassVar[ConfigDict] = cast(
        ConfigDict,
        {
            "populate_by_name": True,
            "extra": "ignore",
        },
    )

    def model_dump_json(self, **kwargs: Any) -> str:
        """Serialize model to JSON using orjson."""
        return orjson.dumps(self.model_dump(**kwargs)).decode()

    @classmethod
    def model_validate_json(
        cls: type["BaseModel"], json_data: Union[str, bytes], **kwargs: Any
    ) -> "BaseModel":
        """Deserialize model from JSON using orjson."""
        return cls.model_validate(orjson.loads(json_data), **kwargs)


class Domain(BaseModel):
    """Domain information for a concept."""

    id: int = Field(..., description="Domain ID")
    name: str = Field(..., description="Domain name")


class Vocabulary(BaseModel):
    """Vocabulary information for a concept."""

    id: str = Field(..., description="Vocabulary ID")
    name: str = Field(..., description="Vocabulary name")


class ConceptClass(BaseModel):
    """Concept class information."""

    id: str = Field(..., description="Concept class ID")
    name: str = Field(..., description="Concept class name")


class ConceptType(str, Enum):
    """Concept standard type."""

    STANDARD = "Standard"
    CLASSIFICATION = "Classification"
    NON_STANDARD = "Non-standard"


class Concept(BaseModel):
    """Basic concept information returned in search results."""

    id: int = Field(..., description="Concept ID")
    name: str = Field(..., description="Concept name")
    domain: str = Field(..., description="Domain name")
    vocabulary: str = Field(..., description="Vocabulary name")
    className: str = Field(..., description="Concept class name")
    standardConcept: Optional[ConceptType] = Field(
        None, description="Standard concept flag"
    )
    code: str = Field(..., description="Concept code")
    invalidReason: Optional[str] = Field(None, description="Invalid reason")
    score: Optional[float] = Field(None, description="Search score")


class ConceptSearchResponse(BaseModel):
    """Response from the /concepts search endpoint."""

    content: List[Concept] = Field(..., description="List of concept results")
    pageable: Dict[str, Any] = Field(..., description="Pagination information")
    totalElements: Optional[int] = Field(None, description="Total number of results")
    last: Optional[bool] = Field(None, description="Whether this is the last page")
    totalPages: Optional[int] = Field(None, description="Total number of pages")
    sort: Optional[Dict[str, Any]] = Field(None, description="Sort information")
    first: Optional[bool] = Field(None, description="Whether this is the first page")
    size: Optional[int] = Field(None, description="Page size")
    number: Optional[int] = Field(None, description="Page number")
    numberOfElements: Optional[int] = Field(
        None, description="Number of elements in this page"
    )
    empty: Optional[bool] = Field(None, description="Whether the result is empty")
    facets: Optional[Dict[str, Any]] = Field(None, description="Search facets")


class ConceptDetails(BaseModel):
    """Detailed concept information from the /concepts/{id} endpoint."""

    id: int = Field(..., description="Concept ID")
    name: str = Field(..., description="Concept name")
    domainId: str = Field(..., description="Domain ID")
    vocabularyId: str = Field(..., description="Vocabulary ID")
    conceptClassId: str = Field(..., description="Concept class ID")
    standardConcept: Optional[ConceptType] = Field(
        None, description="Standard concept flag"
    )
    conceptCode: str = Field(..., description="Concept code")
    invalidReason: Optional[str] = Field(None, description="Invalid reason")
    validStart: str = Field(..., description="Valid start date")
    validEnd: str = Field(..., description="Valid end date")
    synonyms: Optional[List[Union[str, Dict[str, Any]]]] = Field(
        None, description="Concept synonyms"
    )
    validTerm: Optional[Union[str, Dict[str, Any]]] = Field(
        None, description="Valid term"
    )
    vocabularyName: Optional[str] = Field(None, description="Vocabulary name")
    vocabularyVersion: Optional[str] = Field(None, description="Vocabulary version")
    vocabularyReference: Optional[str] = Field(None, description="Vocabulary reference")
    links: Optional[Dict[str, Any]] = Field(
        None, description="HATEOAS links", alias="_links"
    )

    @model_validator(mode="before")
    @classmethod
    def normalize_synonyms(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        synonyms = values.get("synonyms")
        if synonyms and isinstance(synonyms, list):
            normalized = []
            for item in synonyms:
                if isinstance(item, str):
                    normalized.append(item)
                elif isinstance(item, dict):
                    # Use 'synonymName' if present, else join all string values
                    if "synonymName" in item:
                        normalized.append(item["synonymName"])
                    else:
                        # Fallback: join all string values in the dict
                        str_values = [
                            str(v) for v in item.values() if isinstance(v, str)
                        ]
                        normalized.append(", ".join(str_values))
            values["synonyms"] = normalized
        return values

    @model_validator(mode="before")
    @classmethod
    def normalize_valid_term(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        valid_term = values.get("validTerm")
        if valid_term and isinstance(valid_term, dict):
            # Extract name from validTerm dict if it's a dictionary
            if "name" in valid_term:
                values["validTerm"] = valid_term["name"]
            else:
                # Fallback: convert to string representation
                values["validTerm"] = str(valid_term)
        return values


class RelationshipItem(BaseModel):
    """Information about a relationship between concepts."""

    targetConceptId: int = Field(..., description="Target concept ID")
    targetConceptName: str = Field(..., description="Target concept name")
    targetVocabularyId: str = Field(..., description="Target vocabulary ID")
    relationshipId: str = Field(..., description="Relationship ID")
    relationshipName: str = Field(..., description="Relationship name")


class RelationshipGroup(BaseModel):
    """Group of relationships with the same type."""

    relationshipName: str = Field(..., description="Relationship name")
    relationships: List[RelationshipItem] = Field(
        ..., description="List of relationships"
    )


class ConceptRelationship(BaseModel):
    """Response from the /concepts/{id}/relationships endpoint."""

    count: int = Field(..., description="Total count of relationships")
    items: List[RelationshipGroup] = Field(
        ..., description="List of relationship groups"
    )


class GraphTerm(BaseModel):
    """Term in the concept relationship graph."""

    id: int = Field(..., description="Term ID")
    name: str = Field(..., description="Term name")
    weight: int = Field(..., description="Term weight")
    depth: int = Field(..., description="Term depth")
    count: int = Field(..., description="Term count")
    isCurrent: bool = Field(..., description="Whether this is the current concept")


class GraphLink(BaseModel):
    """Link in the concept relationship graph."""

    source: int = Field(..., description="Source term ID")
    target: int = Field(..., description="Target term ID")
    relationshipId: Optional[str] = Field(None, description="Relationship ID")
    relationshipName: Optional[str] = Field(None, description="Relationship name")


class ConceptRelationsGraph(BaseModel):
    """Response from the /concepts/{id}/relations endpoint."""

    terms: List[GraphTerm] = Field(..., description="Graph terms")
    links: List[GraphLink] = Field(..., description="Graph links")
    connectionsCount: Optional[int] = Field(None, description="Total connections count")


# Re-export models
__all__ = [
    "Concept",
    "ConceptClass",
    "ConceptDetails",
    "ConceptRelationsGraph",
    "ConceptRelationship",
    "ConceptSearchResponse",
    "ConceptType",
    "Domain",
    "GraphLink",
    "GraphTerm",
    "RelationshipGroup",
    "RelationshipItem",
    "Vocabulary",
]



"""
Search result wrapper for Athena API responses.

This module provides a wrapper around search results that provides
convenient access to the data in various formats.
"""

from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional

import pandas as pd

from .models import Concept, ConceptSearchResponse

if TYPE_CHECKING:
    pass


class SearchResult:
    """Wrapper for search results that provides convenient access methods."""

    def __init__(self, response: ConceptSearchResponse, client: Any) -> None:
        """Initialize the search result wrapper.

        Args:
            response: The search response from the API
            client: The client instance for making additional requests
        """
        self._response = response
        self._client = client

    def all(self) -> List[Concept]:
        """Get all concepts from the current page.

        Returns:
            List of Concept objects
        """
        return self._response.content

    def top(self, n: int) -> List[Concept]:
        """Get the top N concepts from the current page.

        Args:
            n: Number of concepts to return

        Returns:
            List of Concept objects
        """
        return self._response.content[:n]

    def to_list(self) -> List[Dict[str, Any]]:
        """Convert results to a list of dictionaries.

        Returns:
            List of dictionaries representing concepts
        """
        return [concept.model_dump() for concept in self._response.content]

    def to_json(self) -> str:
        """Convert results to JSON string.

        Returns:
            JSON string representation of the results
        """
        return self._response.model_dump_json()

    def to_df(self) -> pd.DataFrame:
        """Convert results to a pandas DataFrame.

        Returns:
            DataFrame containing the concept data
        """
        try:
            data = self.to_list()
            return pd.DataFrame(data)
        except ImportError:
            raise ImportError(
                "pandas is required for DataFrame output. "
                "Install with: pip install 'athena-client[pandas]'"
            ) from None

    def next_page(self) -> Optional["SearchResult"]:
        """Get the next page of results.

        Returns:
            SearchResult for the next page, or None if no more pages
        """
        if self._response.last:
            return None
        current_page = self._response.number
        size = self._response.size
        if current_page is None or size is None:
            return None
        return self._client.search(
            query="",  # This would need to be the original query
            page=current_page + 1,
            size=size,
        )

    def previous_page(self) -> Optional["SearchResult"]:
        """Get the previous page of results.

        Returns:
            SearchResult for the previous page, or None if no previous pages
        """
        if self._response.first:
            return None
        current_page = self._response.number
        size = self._response.size
        if current_page is None or size is None:
            return None
        return self._client.search(
            query="",  # This would need to be the original query
            page=current_page - 1,
            size=size,
        )

    @property
    def total_elements(self) -> int:
        """Get the total number of elements across all pages.

        Returns:
            Total number of elements
        """
        # Try to get from direct field first, then from pageable
        if self._response.totalElements is not None:
            return self._response.totalElements

        # Extract from pageable if available
        pageable = self._response.pageable
        if pageable and "totalElements" in pageable:
            return pageable["totalElements"]

        # Fallback to number of elements in current page
        return len(self._response.content)

    @property
    def total_pages(self) -> int:
        """Get the total number of pages.

        Returns:
            Total number of pages
        """
        # Try to get from direct field first, then calculate from pageable
        if self._response.totalPages is not None:
            return self._response.totalPages

        # Calculate from pageable if available
        pageable = self._response.pageable
        if pageable and "totalElements" in pageable and "pageSize" in pageable:
            total_elements = pageable["totalElements"]
            page_size = pageable["pageSize"]
            return (total_elements + page_size - 1) // page_size

        return 1

    @property
    def current_page(self) -> int:
        """Get the current page number.

        Returns:
            Current page number
        """
        # Try to get from direct field first, then from pageable
        if self._response.number is not None:
            return self._response.number

        # Extract from pageable if available
        pageable = self._response.pageable
        if pageable and "pageNumber" in pageable:
            return pageable["pageNumber"]

        return 0

    @property
    def page_size(self) -> int:
        """Get the page size.

        Returns:
            Page size
        """
        # Try to get from direct field first, then from pageable
        if self._response.size is not None:
            return self._response.size

        # Extract from pageable if available
        pageable = self._response.pageable
        if pageable and "pageSize" in pageable:
            return pageable["pageSize"]

        return len(self._response.content)

    @property
    def facets(self) -> Optional[Dict[str, Any]]:
        """Get search facets if available.

        Returns:
            Search facets dictionary or None
        """
        return self._response.facets

    def __len__(self) -> int:
        """Get the number of concepts in the current page.

        Returns:
            Number of concepts
        """
        return len(self._response.content)

    def __getitem__(self, index: int) -> Concept:
        """Get a concept by index.

        Args:
            index: Index of the concept

        Returns:
            Concept object
        """
        return self._response.content[index]

    def __iter__(self) -> Iterator["Concept"]:
        """Iterate over concepts in the current page.

        Returns:
            Iterator over Concept objects
        """
        return iter(self._response.content)



"""
Authentication module for the Athena client.

This module handles Bearer token and HMAC authentication for the Athena API.
"""

import logging
from base64 import b64encode
from datetime import datetime
from typing import Any, Dict

from .settings import get_settings

logger = logging.getLogger(__name__)


def build_headers(
    method: str,
    url: str,
    body: bytes,
    serialization_module: Any = None,
    hashes_module: Any = None,
) -> Dict[str, str]:
    """
    Build authentication headers for Athena API requests.

    If a token is supplied, adds Bearer authentication.
    If a private key is supplied, adds HMAC signature.

    Args:
        method: HTTP method (GET, POST, etc.)
        url: Full request URL
        body: Request body as bytes

    Returns:
        Dictionary of headers to add to the request
    """
    s = get_settings()

    # Start with empty headers - authentication is optional
    hdrs = {}

    # Add Bearer token if available
    if s.ATHENA_TOKEN:
        hdrs["X-Athena-Auth"] = f"Bearer {s.ATHENA_TOKEN}"
        hdrs["X-Athena-Client-Id"] = s.ATHENA_CLIENT_ID or "athena-client"
        logger.debug("Bearer token authentication headers added")
    else:
        logger.debug("No API token provided; proceeding without Authorization header")

    # Add HMAC signature if private key is available
    if s.ATHENA_PRIVATE_KEY:
        try:
            if serialization_module is None or hashes_module is None:
                from cryptography.hazmat.primitives import hashes, serialization

                serialization_module = serialization
                hashes_module = hashes
        except ImportError:
            logger.warning(
                "cryptography package is required for HMAC signing. "
                "Install with 'pip install \"athena-client[crypto]\"'"
            )
            return hdrs
        try:
            nonce = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")
            to_sign = f"{method}\n{url}\n\n{nonce}\n{body.decode()}"
            key = serialization_module.load_pem_private_key(
                s.ATHENA_PRIVATE_KEY.encode(), password=None
            )
            signing_key: Any = key
            sig = signing_key.sign(to_sign.encode(), hashes_module.SHA256())
            hdrs.update(
                {"X-Athena-Nonce": nonce, "X-Athena-Hmac": b64encode(sig).decode()}
            )
            logger.debug("HMAC signature headers added")
        except Exception as e:
            logger.error(f"Error generating HMAC signature: {e}")

    return hdrs



"""
Utility functions for the Athena client.

This module provides various utility functions for the Athena client,
including progress tracking, query size estimation, and timeout management.
"""

import logging
from typing import Optional

from .progress import (
    ProgressTracker,
    estimate_query_size,
    format_large_query_warning,
    get_operation_timeout,
    progress_context,
)

logger = logging.getLogger(__name__)


def configure_logging(level: Optional[int] = None) -> None:
    """
    Configure logging for the Athena client.

    Args:
        level: Logging level (e.g., logging.INFO, logging.DEBUG)
    """
    log_level = level or logging.INFO

    logger = logging.getLogger("athena_client")
    logger.setLevel(log_level)

    # Create console handler if no handlers exist
    if not logger.handlers:
        handler = logging.StreamHandler()
        handler.setLevel(log_level)

        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)

        logger.addHandler(handler)


__all__ = [
    "configure_logging",
    "ProgressTracker",
    "progress_context",
    "estimate_query_size",
    "get_operation_timeout",
    "format_large_query_warning",
]



"""
Tests for exception classes and error handling.
"""

from athena_client.exceptions import (
    APIError,
    AthenaError,
    AuthenticationError,
    ClientError,
    NetworkError,
    RateLimitError,
    RetryFailedError,
    ServerError,
    TimeoutError,
    ValidationError,
)


class TestAthenaError:
    """Test base AthenaError class."""

    def test_athena_error_initialization(self):
        """Test AthenaError initialization."""
        error = AthenaError("Test error message")
        assert error.message == "Test error message"
        assert error.error_code is None  # Default value
        assert error.troubleshooting is None  # Default value

    def test_athena_error_string_representation(self):
        """Test AthenaError string representation."""
        error = AthenaError("Test error message")
        error_str = str(error)
        assert "Test error message" in error_str
        # No troubleshooting since it's None

    def test_athena_error_with_custom_values(self):
        """Test AthenaError with custom error_code and troubleshooting."""
        error = AthenaError(
            "Test error message",
            error_code="TEST_ERROR",
            troubleshooting="• Step 1\n• Step 2",
        )
        assert error.message == "Test error message"
        assert error.error_code == "TEST_ERROR"
        assert error.troubleshooting == "• Step 1\n• Step 2"

    def test_athena_error_string_with_troubleshooting(self):
        """Test AthenaError string representation with troubleshooting."""
        error = AthenaError("Test error message", troubleshooting="• Step 1\n• Step 2")
        error_str = str(error)
        assert "Test error message" in error_str
        assert "Step 1" in error_str
        assert "Step 2" in error_str

    def test_athena_error_default_values(self):
        """Test AthenaError default values."""
        error = AthenaError("Simple error")
        assert error.error_code is None  # Default value
        assert error.troubleshooting is None  # Default value


class TestNetworkError:
    """Test NetworkError class."""

    def test_network_error_initialization(self):
        """Test NetworkError initialization."""
        error = NetworkError("Connection failed")
        assert error.message == "Connection failed"
        assert error.error_code == "NETWORK_ERROR"
        assert "internet connection" in error.troubleshooting.lower()

    def test_network_error_with_url(self):
        """Test NetworkError with URL context."""
        error = NetworkError("Connection failed", url="https://api.example.com/test")
        assert "Connection failed" in str(error)

    def test_network_error_message(self):
        err = NetworkError("Connection failed")
        assert "Connection failed" in str(err)


class TestTimeoutError:
    """Test TimeoutError class."""

    def test_timeout_error_initialization(self):
        """Test TimeoutError initialization."""
        error = TimeoutError("Request timeout")
        assert error.message == "Request timeout"
        assert error.error_code == "NETWORK_ERROR"
        assert "timeout" in error.troubleshooting.lower()

    def test_timeout_error_with_timeout_value(self):
        """Test TimeoutError with timeout value."""
        error = TimeoutError("Request timeout")
        assert "Request timeout" in str(error)

    def test_timeout_error_message(self):
        err = TimeoutError("Request timed out")
        assert "Request timed out" in str(err)


class TestValidationError:
    """Test ValidationError class."""

    def test_validation_error_initialization(self):
        """Test ValidationError initialization."""
        error = ValidationError("Invalid JSON response")
        assert error.message == "Invalid JSON response"
        assert error.error_code == "VALIDATION_ERROR"
        assert "api response format" in error.troubleshooting.lower()

    def test_validation_error_string_representation(self):
        """Test ValidationError string representation."""
        error = ValidationError("Invalid JSON response")
        error_str = str(error)
        assert "Invalid JSON response" in error_str

    def test_validation_error_message(self):
        err = ValidationError("Invalid data")
        assert "Invalid data" in str(err)


class TestAuthenticationError:
    """Test AuthenticationError class."""

    def test_authentication_error_initialization(self):
        """Test AuthenticationError initialization."""
        error = AuthenticationError("Unauthorized access", status_code=403)
        assert error.message == "Unauthorized access"
        assert error.error_code == "AUTHENTICATION_ERROR"
        assert error.status_code == 403
        assert "authentication" in error.troubleshooting.lower()

    def test_authentication_error_string_representation(self):
        """Test AuthenticationError string representation."""
        error = AuthenticationError("Unauthorized access", status_code=403)
        error_str = str(error)
        assert "Unauthorized access" in error_str

    def test_authentication_error_message(self):
        err = AuthenticationError("Unauthorized", status_code=401)
        assert "Unauthorized" in str(err)


class TestRateLimitError:
    """Test RateLimitError class."""

    def test_rate_limit_error_initialization(self):
        """Test RateLimitError initialization."""
        error = RateLimitError("Rate limit exceeded", status_code=429)
        assert error.message == "Rate limit exceeded"
        assert error.error_code == "RATE_LIMIT_ERROR"
        assert error.status_code == 429
        assert "rate limit" in error.troubleshooting.lower()

    def test_rate_limit_error_string_representation(self):
        """Test RateLimitError string representation."""
        error = RateLimitError("Rate limit exceeded", status_code=429)
        error_str = str(error)
        assert "Rate limit exceeded" in error_str

    def test_rate_limit_error_message(self):
        err = RateLimitError("Rate limit exceeded", status_code=429)
        assert "Rate limit exceeded" in str(err)


class TestClientError:
    """Test ClientError class."""

    def test_client_error_initialization(self):
        """Test ClientError initialization."""
        error = ClientError("Bad request", status_code=400)
        assert error.message == "Bad request"
        assert error.error_code == "CLIENT_ERROR"
        assert error.status_code == 400
        assert "request parameters" in error.troubleshooting.lower()

    def test_client_error_string_representation(self):
        """Test ClientError string representation."""
        error = ClientError("Invalid parameters", status_code=422)
        error_str = str(error)
        assert "Invalid parameters" in error_str

    def test_client_error_message(self):
        err = ClientError("Bad request", status_code=400)
        assert "Bad request" in str(err)


class TestServerError:
    """Test ServerError class."""

    def test_server_error_initialization(self):
        """Test ServerError initialization."""
        error = ServerError("Service unavailable", status_code=503)
        assert error.message == "Service unavailable"
        assert error.error_code == "SERVER_ERROR"
        assert error.status_code == 503
        assert "server" in error.troubleshooting.lower()

    def test_server_error_string_representation(self):
        """Test ServerError string representation."""
        error = ServerError("Service unavailable", status_code=503)
        error_str = str(error)
        assert "Service unavailable" in error_str

    def test_server_error_message(self):
        err = ServerError("Internal error", status_code=500)
        assert "Internal error" in str(err)


class TestAPIError:
    """Test APIError class."""

    def test_api_error_initialization(self):
        """Test APIError initialization."""
        error = APIError("Search failed", api_error_code="INVALID_QUERY")
        assert error.message == "Search failed"
        assert error.error_code == "API_ERROR"
        assert error.api_error_code == "INVALID_QUERY"

    def test_api_error_string_representation(self):
        """Test APIError string representation."""
        error = APIError("Search failed", api_error_code="INVALID_QUERY")
        error_str = str(error)
        assert "Search failed" in error_str

    def test_api_error_message(self):
        err = APIError("Concept not found", api_error_code="NOT_FOUND")
        assert "Concept not found" in str(err)


class TestRetryFailedError:
    """Test RetryFailedError class."""

    def test_retry_failed_error_initialization(self):
        """Test RetryFailedError initialization."""
        last_error = NetworkError("Connection failed")
        error = RetryFailedError(
            "Retry failed after 3 attempts",
            max_attempts=3,
            last_error=last_error,
            retry_history=[last_error, last_error, last_error],
        )

        assert error.message == "Retry failed after 3 attempts"
        assert error.error_code == "RETRY_FAILED"
        assert error.max_attempts == 3
        assert error.last_error == last_error

    def test_retry_failed_error_message(self):
        err = RetryFailedError(
            "Retry failed after 3 attempts",
            max_attempts=3,
            last_error="Timeout",
            retry_history=["Timeout", "Timeout", "Timeout"],
        )
        assert "Retry failed after 3 attempts" in str(err)
        assert "3" in str(err)
        assert "Timeout" in str(err)


class TestExceptionInheritance:
    """Test exception inheritance hierarchy."""

    def test_exception_inheritance(self):
        """Test that all exceptions inherit from AthenaError."""
        exceptions = [
            NetworkError("test"),
            TimeoutError("test"),
            ValidationError("test"),
            AuthenticationError("test", status_code=401),
            RateLimitError("test", status_code=429),
            ClientError("test", status_code=400),
            ServerError("test", status_code=500),
            APIError("test"),
            RetryFailedError("test", 1, NetworkError("test"), []),
        ]

        for exc in exceptions:
            assert hasattr(exc, "message")
            assert hasattr(exc, "error_code")
            assert hasattr(exc, "troubleshooting")

    def test_exception_types(self):
        """Test that exceptions are of correct types."""
        assert isinstance(NetworkError("test"), NetworkError)
        assert isinstance(TimeoutError("test"), TimeoutError)
        assert isinstance(ValidationError("test"), ValidationError)
        assert isinstance(
            AuthenticationError("test", status_code=401), AuthenticationError
        )
        assert isinstance(RateLimitError("test", status_code=429), RateLimitError)
        assert isinstance(ClientError("test", status_code=400), ClientError)
        assert isinstance(ServerError("test", status_code=500), ServerError)
        assert isinstance(APIError("test"), APIError)
        assert isinstance(
            RetryFailedError("test", 1, NetworkError("test"), []), RetryFailedError
        )


class TestExceptionContext:
    """Test exception context and troubleshooting."""

    def test_network_error_context(self):
        """Test NetworkError provides network-specific troubleshooting."""
        error = NetworkError("Connection failed")
        assert "internet connection" in error.troubleshooting.lower()
        assert "network" in error.troubleshooting.lower()

    def test_authentication_error_context(self):
        """Test AuthenticationError provides auth-specific troubleshooting."""
        error = AuthenticationError("Invalid token")
        assert "token" in error.troubleshooting.lower()
        assert "authentication" in error.troubleshooting.lower()

    def test_rate_limit_error_context(self):
        """Test RateLimitError provides rate limit-specific troubleshooting."""
        error = RateLimitError("Rate limit exceeded", status_code=429)
        assert "rate limit" in error.troubleshooting.lower()
        assert "wait" in error.troubleshooting.lower()

    def test_validation_error_context(self):
        """Test ValidationError provides validation-specific troubleshooting."""
        error = ValidationError("Invalid data")
        assert "api response format" in error.troubleshooting.lower()

    def test_retry_failed_error_context(self):
        """Test RetryFailedError provides retry-specific troubleshooting."""
        error = RetryFailedError("Failed", 3, NetworkError("test"), [])
        assert "retry attempts have been exhausted" in error.troubleshooting
        assert "max_retries" in error.troubleshooting.lower()



"""
Main client for the Athena API.

This module provides the main client class for interacting with the Athena API.
"""

import logging
import time
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

from .async_client import AthenaAsyncClient
from .db.base import DatabaseConnector
from .db.sqlalchemy_connector import SQLAlchemyConnector
from .exceptions import APIError, AthenaError
from .http import HttpClient
from .models import (
    ConceptDetails,
    ConceptRelationsGraph,
    ConceptRelationship,
    ConceptSearchResponse,
)
from .query import Q
from .search_result import SearchResult
from .settings import get_settings

logger = logging.getLogger(__name__)


class AthenaClient:
    """Main client for interacting with the Athena API."""

    def __init__(
        self,
        base_url: Optional[str] = None,
        token: Optional[str] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        retry_delay: Optional[float] = None,
        enable_throttling: bool = True,
        throttle_delay_range: Tuple[float, float] = (0.1, 0.3),
        db_connector: Optional[DatabaseConnector] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Athena client.

        Args:
            base_url: Base URL for the Athena API
            token: Authentication token
            timeout: Request timeout in seconds
            max_retries: Maximum number of retries for network errors
            retry_delay: Delay between retry attempts in seconds
                (overrides backoff_factor)
            enable_throttling: Whether to throttle requests to be respectful
                to the server
            throttle_delay_range: Range of delays for throttling (min, max)
                in seconds
            db_connector: Optional database connector for local OMOP validation
            **kwargs: Additional settings
        """
        s = get_settings()

        # Configure retry settings
        self.max_retries = max_retries or s.ATHENA_MAX_RETRIES
        self.retry_delay = retry_delay
        self.enable_throttling = enable_throttling
        self.throttle_delay_range = throttle_delay_range
        self._db_connector = db_connector

        # Create HTTP client with enhanced configuration
        self.http = HttpClient(
            base_url=base_url or s.ATHENA_BASE_URL,
            token=token or s.ATHENA_TOKEN,
            timeout=timeout or s.ATHENA_TIMEOUT_SECONDS,
            max_retries=self.max_retries,
            enable_throttling=self.enable_throttling,
            throttle_delay_range=self.throttle_delay_range,
            **kwargs,
        )

        logger.info(
            f"Athena client initialized: "
            f"max_retries={self.max_retries}, "
            f"retry_delay={self.retry_delay}, "
            f"throttling={'enabled' if self.enable_throttling else 'disabled'}"
        )

    def set_database_connector(self, connector: DatabaseConnector) -> None:
        """Set the database connector for this client."""

        self._db_connector = connector

    def validate_local_concepts(self, concept_ids: List[int]) -> List[int]:
        """Validate concept IDs against the configured local database."""

        if not self._db_connector:
            raise RuntimeError(
                "A database connector has not been configured. Use "
                "`set_database_connector()` to provide one."
            )

        return self._db_connector.validate_concepts(concept_ids)

    def search(
        self,
        query: Union[str, Q],
        page: int = 0,
        size: int = 20,
        auto_retry: bool = True,
        max_retries: Optional[int] = None,
        retry_delay: Optional[float] = None,
        show_progress: Optional[bool] = None,
        **kwargs: Any,
    ) -> SearchResult:
        """
        Search for concepts with automatic error handling and recovery.

        Args:
            query: Search query string or Query DSL object
            page: Page number (0-based)
            size: Page size
            auto_retry: Whether to automatically retry on recoverable errors
            max_retries: Override max retries for this call
            retry_delay: Override retry delay for this call
            show_progress: Whether to show progress for large queries
            **kwargs: Additional search parameters

        Returns:
            SearchResult object with concepts

        Note:
            This method includes automatic error handling and recovery.
            Network errors are automatically retried, and API errors provide
            clear, actionable messages without requiring try-catch blocks.
        """

        from .exceptions import RetryFailedError
        from .settings import get_settings
        from .utils import (
            estimate_query_size,
            format_large_query_warning,
            get_operation_timeout,
            progress_context,
        )

        settings = get_settings()

        if isinstance(query, Q):
            query_str = str(query)
        else:
            query_str = query

        # Estimate query size and provide warnings for large queries
        estimated_size = estimate_query_size(query_str)
        warning = format_large_query_warning(query_str, estimated_size)
        if warning:
            print(warning)

        # Use settings default page size if not specified
        if size == 20 and "pageSize" not in kwargs:
            size = settings.ATHENA_DEFAULT_PAGE_SIZE

        # Validate page size
        if size > settings.ATHENA_MAX_PAGE_SIZE:
            raise ValueError(
                f"Page size {size} exceeds maximum allowed size of "
                f"{settings.ATHENA_MAX_PAGE_SIZE}. Please use a smaller page size."
            )

        # Convert page/size to pageSize/start parameters that the API expects
        page_size = size
        start = page * size  # Calculate start offset

        params = {
            "query": query_str,
            "pageSize": page_size,
            "start": start,
            **kwargs,
        }

        # Configure retry settings for this call
        max_attempts = max(
            1,
            max_retries
            if max_retries is not None
            else (self.max_retries if auto_retry else 1),
        )

        # Get appropriate timeout for this operation
        operation_timeout = get_operation_timeout("search", estimated_size)

        # Set up progress tracking if enabled
        progress_kwargs: Optional[Dict[str, Any]] = None
        if show_progress:
            progress_kwargs = {
                "total": estimated_size,
                "description": f"Searching for '{query_str}'",
                "show_progress": show_progress,
                "update_interval": settings.ATHENA_PROGRESS_UPDATE_INTERVAL,
            }

        retry_history: list[Exception] = []
        with progress_context(**progress_kwargs) if progress_kwargs else nullcontext():
            for attempt in range(max_attempts):
                try:
                    # Create a temporary HTTP client with the appropriate timeout
                    temp_http = HttpClient(
                        base_url=self.http.base_url,
                        token=str(self.http.session.headers.get("Authorization", "")),
                        timeout=operation_timeout,
                        max_retries=1,  # We handle retries manually
                        enable_throttling=self.http.enable_throttling,
                        throttle_delay_range=self.http.throttle_delay_range,
                    )

                    response = temp_http.get("/concepts", params=params)

                    # Raise APIError for any error response with errorMessage
                    # and errorCode
                    if (
                        isinstance(response, dict)
                        and response.get("result") is None
                        and "errorMessage" in response
                        and "errorCode" in response
                    ):
                        error_msg = response.get("errorMessage", "Unknown API error")
                        error_code = response.get("errorCode")

                        # Enhanced error messages for large queries
                        if "timeout" in error_msg.lower():
                            raise APIError(
                                f"Search timeout: The query '{query_str}' "
                                f"is taking too long to process. "
                                f"Try:\n"
                                f"• Using more specific search terms\n"
                                f"• Adding domain or vocabulary filters\n"
                                f"• Reducing the page size\n"
                                f"• Breaking the query into smaller parts",
                                api_error_code=error_code,
                                api_message=error_msg,
                            )
                        elif "Page size must not be less than one" in error_msg:
                            raise APIError(
                                f"Invalid page size: {error_msg}. "
                                f"Please use a page size of 1 or greater.",
                                api_error_code=error_code,
                                api_message=error_msg,
                            )
                        elif "Page size must not be greater than" in error_msg:
                            raise APIError(
                                f"Page size too large: {error_msg}. "
                                f"Please reduce the page size to "
                                f"{settings.ATHENA_MAX_PAGE_SIZE} or less.",
                                api_error_code=error_code,
                                api_message=error_msg,
                            )
                        elif "Query must not be blank" in error_msg:
                            raise APIError(
                                f"Empty search query: {error_msg}. "
                                f"Please provide a search term.",
                                api_error_code=error_code,
                                api_message=error_msg,
                            )
                        else:
                            raise APIError(
                                f"Search failed: {error_msg}",
                                api_error_code=error_code,
                                api_message=error_msg,
                            )

                    search_response = ConceptSearchResponse.model_validate(response)
                    return SearchResult(search_response, self)

                except Exception as e:
                    if isinstance(e, APIError):
                        # API errors are not retryable, raise immediately
                        raise

                    # For network errors, retry if we have attempts left
                    if attempt < max_attempts - 1:
                        logger.info(
                            f"Retrying search due to {type(e).__name__} "
                            f"(attempt {attempt + 1}/{max_attempts}): {e}"
                        )
                        if retry_delay is not None:
                            time.sleep(retry_delay)
                        elif self.retry_delay is not None:
                            time.sleep(self.retry_delay)
                        retry_history.append(e)
                        continue
                    else:
                        # Final attempt failed, raise with retry history
                        raise RetryFailedError(
                            f"Search failed after {max_attempts} attempts",
                            retry_history=retry_history,
                            max_attempts=max_attempts,
                            last_error=e,
                        ) from e
            raise RuntimeError("Unreachable code in search")

    def details(self, concept_id: int, auto_retry: bool = True) -> ConceptDetails:
        """
        Get detailed information about a concept with automatic error handling.

        Args:
            concept_id: Concept ID
            auto_retry: Whether to automatically retry on recoverable errors

        Returns:
            ConceptDetails object

        Note:
            This method includes automatic error handling and recovery.
            Network errors are automatically retried, and API errors provide
            clear, actionable messages without requiring try-catch blocks.
        """
        max_attempts = 3 if auto_retry else 1

        for attempt in range(max_attempts):
            try:
                response = self.http.get(f"/concepts/{concept_id}")

                # Check if the response is an error response
                if (
                    isinstance(response, dict)
                    and response.get("result") is None
                    and "errorMessage" in response
                ):
                    error_msg = response.get("errorMessage", "Unknown API error")
                    error_code = response.get("errorCode")

                    # Provide more specific error messages for concept details
                    if "Unable to find" in error_msg and "ConceptV5" in error_msg:
                        raise APIError(
                            f"Concept not found: Concept ID {concept_id} "
                            f"does not exist in the database. "
                            f"Please verify the concept ID is correct.",
                            api_error_code=error_code,
                            api_message=error_msg,
                        )
                    else:
                        raise APIError(
                            f"Failed to get concept details: {error_msg}",
                            api_error_code=error_code,
                            api_message=error_msg,
                        )

                return ConceptDetails.model_validate(response)

            except Exception as e:
                if isinstance(e, APIError):
                    # API errors are not retryable, raise immediately
                    raise
                elif attempt < max_attempts - 1:
                    # For other errors, retry if we have attempts left
                    logger.info(
                        f"Retrying concept details due to error "
                        f"(attempt {attempt + 1}/{max_attempts}): {e}"
                    )
                    continue
                else:
                    # Final attempt failed, raise with enhanced message
                    raise AthenaError(
                        f"Failed to get concept details after {max_attempts} attempts. "
                        f"Last error: {e}",
                        error_code="RETRY_FAILED",
                        troubleshooting=(
                            "• Check your internet connection\n"
                            "• Try again in a few moments\n"
                            "• Contact support if the problem persists"
                        ),
                    ) from e
        raise RuntimeError("Unreachable code in details")

    def relationships(
        self, concept_id: int, auto_retry: bool = True
    ) -> ConceptRelationship:
        """
        Get relationships for a concept with automatic error handling.

        Args:
            concept_id: Concept ID
            auto_retry: Whether to automatically retry on recoverable errors

        Returns:
            ConceptRelationship object

        Note:
            This method includes automatic error handling and recovery.
            Network errors are automatically retried, and API errors provide
            clear, actionable messages without requiring try-catch blocks.
        """
        max_attempts = 3 if auto_retry else 1

        for attempt in range(max_attempts):
            try:
                response = self.http.get(f"/concepts/{concept_id}/relationships")

                # Check if the response is an error response
                if (
                    isinstance(response, dict)
                    and response.get("result") is None
                    and "errorMessage" in response
                ):
                    error_msg = response.get("errorMessage", "Unknown API error")
                    error_code = response.get("errorCode")

                    # Provide more specific error messages for relationships
                    if "Unable to find" in error_msg and "ConceptV5" in error_msg:
                        raise APIError(
                            f"Concept not found: Concept ID {concept_id} "
                            f"does not exist in the database. "
                            f"Please verify the concept ID is correct.",
                            api_error_code=error_code,
                            api_message=error_msg,
                        )
                    else:
                        raise APIError(
                            f"Failed to get relationships: {error_msg}",
                            api_error_code=error_code,
                            api_message=error_msg,
                        )

                return ConceptRelationship.model_validate(response)

            except Exception as e:
                if isinstance(e, APIError):
                    # API errors are not retryable, raise immediately
                    raise
                elif attempt < max_attempts - 1:
                    # For other errors, retry if we have attempts left
                    logger.info(
                        f"Retrying relationships due to error "
                        f"(attempt {attempt + 1}/{max_attempts}): {e}"
                    )
                    continue
                else:
                    # Final attempt failed, raise with enhanced message
                    raise AthenaError(
                        f"Failed to get relationships after {max_attempts} attempts. "
                        f"Last error: {e}",
                        error_code="RETRY_FAILED",
                        troubleshooting=(
                            "• Check your internet connection\n"
                            "• Try again in a few moments\n"
                            "• Contact support if the problem persists"
                        ),
                    ) from e
        raise RuntimeError("Unreachable code in relationships")

    def graph(
        self,
        concept_id: int,
        depth: int = 2,
        zoom_level: int = 2,
        auto_retry: bool = True,
        show_progress: Optional[bool] = None,
        **kwargs: Any,
    ) -> ConceptRelationsGraph:
        """
        Get concept relationship graph with automatic error handling.

        Args:
            concept_id: Concept ID
            depth: Graph depth
            zoom_level: Zoom level
            auto_retry: Whether to automatically retry on recoverable errors
            show_progress: Whether to show progress for large graph operations
            **kwargs: Additional parameters

        Returns:
            ConceptRelationsGraph object

        Note:
            This method includes automatic error handling and recovery.
            Network errors are automatically retried, and API errors provide
            clear, actionable messages without requiring try-catch blocks.
        """
        from .settings import get_settings
        from .utils import get_operation_timeout, progress_context

        settings = get_settings()

        # Estimate graph complexity based on depth and zoom level
        estimated_complexity = depth * zoom_level * 100  # Rough estimate

        # Provide warning for complex graphs
        if depth > 3 or zoom_level > 3:
            print(f"⚠️  Complex graph request: depth={depth}, zoom_level={zoom_level}")
            print(
                "💡 This may take several minutes to complete. "
                "Consider reducing depth or zoom level."
            )

        params = {
            "depth": depth,
            "zoomLevel": zoom_level,
            **kwargs,
        }

        max_attempts = 3 if auto_retry else 1

        # Get appropriate timeout for graph operations
        operation_timeout = get_operation_timeout("graph", estimated_complexity)

        # Set up progress tracking if enabled
        progress_kwargs: Optional[Dict[str, Any]] = None
        if show_progress:
            progress_kwargs = {
                "total": estimated_complexity,
                "description": (
                    f"Building graph for concept {concept_id} "
                    f"(depth={depth}, zoom={zoom_level})"
                ),
                "show_progress": show_progress,
                "update_interval": settings.ATHENA_PROGRESS_UPDATE_INTERVAL,
            }

        with progress_context(**progress_kwargs) if progress_kwargs else nullcontext():
            for attempt in range(max_attempts):
                try:
                    # Create a temporary HTTP client with the appropriate timeout
                    temp_http = HttpClient(
                        base_url=self.http.base_url,
                        token=str(self.http.session.headers.get("Authorization", "")),
                        timeout=operation_timeout,
                        max_retries=1,  # We handle retries manually
                        enable_throttling=self.http.enable_throttling,
                        throttle_delay_range=self.http.throttle_delay_range,
                    )

                    response = temp_http.get(
                        f"/concepts/{concept_id}/relations", params=params
                    )

                    # Check if the response is an error response
                    if (
                        isinstance(response, dict)
                        and response.get("result") is None
                        and "errorMessage" in response
                    ):
                        error_msg = response.get("errorMessage", "Unknown API error")
                        error_code = response.get("errorCode")

                        # Enhanced error messages for graph operations
                        if "timeout" in error_msg.lower():
                            raise APIError(
                                f"Graph timeout: The graph for concept {concept_id} "
                                f"is too complex. "
                                f"Try:\n"
                                f"• Reducing the depth (currently {depth})\n"
                                f"• Reducing the zoom level (currently {zoom_level})\n"
                                f"• Using a simpler concept as the starting point\n"
                                f"• Breaking the request into smaller parts",
                                api_error_code=error_code,
                                api_message=error_msg,
                            )
                        elif "Unable to find" in error_msg and "ConceptV5" in error_msg:
                            raise APIError(
                                f"Concept not found: Concept ID {concept_id} "
                                f"does not exist in the database. "
                                f"Please verify the concept ID is correct.",
                                api_error_code=error_code,
                                api_message=error_msg,
                            )
                        else:
                            raise APIError(
                                f"Failed to get concept graph: {error_msg}",
                                api_error_code=error_code,
                                api_message=error_msg,
                            )

                    return ConceptRelationsGraph.model_validate(response)

                except Exception as e:
                    if isinstance(e, APIError):
                        # API errors are not retryable, raise immediately
                        raise
                    elif attempt < max_attempts - 1:
                        # For other errors, retry if we have attempts left
                        logger.info(
                            f"Retrying graph due to error "
                            f"(attempt {attempt + 1}/{max_attempts}): {e}"
                        )
                        continue
                    else:
                        # Final attempt failed, raise with enhanced message
                        raise AthenaError(
                            f"Failed to get concept graph after {max_attempts} "
                            f"attempts. "
                            f"Last error: {e}",
                            error_code="RETRY_FAILED",
                            troubleshooting=(
                                "• Check your internet connection\n"
                                "• Try again in a few moments\n"
                                "• Consider reducing graph depth or zoom level\n"
                                "• Contact support if the problem persists"
                            ),
                        ) from e
            raise RuntimeError("Unreachable code in graph")

    def summary(
        self,
        concept_id: int,
        include_relationships: bool = True,
        include_graph: bool = True,
    ) -> Dict[str, Any]:
        """Get a comprehensive summary of a concept.

        Args:
            concept_id: Concept ID
            include_relationships: Whether to include relationships
            include_graph: Whether to include graph data

        Returns:
            Dictionary containing concept summary

        Raises:
            AthenaError: If any request fails
        """
        summary = {}

        # Get basic details
        try:
            details = self.details(concept_id)
            summary["details"] = details.model_dump()
        except Exception as e:
            summary["details"] = {"error": str(e)}

        # Get relationships if requested
        if include_relationships:
            try:
                relationships = self.relationships(concept_id)
                summary["relationships"] = relationships.model_dump()
            except Exception as e:
                summary["relationships"] = {"error": str(e)}

        # Get graph if requested
        if include_graph:
            try:
                graph = self.graph(concept_id)
                summary["graph"] = graph.model_dump()
            except Exception as e:
                summary["graph"] = {"error": str(e)}

        return summary

    async def generate_concept_set(
        self,
        query: str,
        db_connection_string: str,
        strategy: str = "fallback",
        include_descendants: bool = True,
        confidence_threshold: float = 0.7,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Generate a validated concept set using the async client."""

        async_client = AthenaAsyncClient(
            base_url=self.http.base_url,
            token=str(self.http.session.headers.get("Authorization", "")),
        )

        db_connector = SQLAlchemyConnector.from_connection_string(db_connection_string)
        async_client.set_database_connector(db_connector)

        return await async_client.generate_concept_set(
            query,
            strategy=strategy,
            include_descendants=include_descendants,
            confidence_threshold=confidence_threshold,
            **kwargs,
        )


class Athena(AthenaClient):
    """Alias for AthenaClient for backward compatibility."""

    pass


class nullcontext:
    """Context manager that does nothing."""

    def __enter__(self) -> None:
        return None

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> Literal[False]:
        return False



"""
Settings module for the Athena client.

This module provides configuration settings loaded from environment variables
or .env files using pydantic-settings.
"""

from typing import Optional

from pydantic_settings import BaseSettings, SettingsConfigDict


class _Settings(BaseSettings):
    """
    Configuration settings for the Athena client.

    Settings can be provided via environment variables, .env file, or defaults.
    The settings are designed to be isolated from other applications' environment
    variables and will only use ATHENA_ prefixed variables.
    """

    ATHENA_BASE_URL: str = "https://athena.ohdsi.org/api/v1"
    ATHENA_TOKEN: Optional[str] = None
    ATHENA_CLIENT_ID: Optional[str] = None
    ATHENA_PRIVATE_KEY: Optional[str] = None

    # Enhanced timeout configuration for different operations
    ATHENA_TIMEOUT_SECONDS: int = 30  # Increased default timeout
    ATHENA_SEARCH_TIMEOUT_SECONDS: int = 45  # Longer timeout for search operations
    ATHENA_GRAPH_TIMEOUT_SECONDS: int = 60  # Even longer for graph operations
    ATHENA_RELATIONSHIPS_TIMEOUT_SECONDS: int = 45  # For relationship queries

    # Retry configuration
    ATHENA_MAX_RETRIES: int = 3
    ATHENA_BACKOFF_FACTOR: float = 0.3

    # Pagination configuration
    ATHENA_DEFAULT_PAGE_SIZE: int = 20
    ATHENA_MAX_PAGE_SIZE: int = 1000
    ATHENA_LARGE_QUERY_THRESHOLD: int = 100  # Threshold for "large" queries

    # Progress and user experience
    ATHENA_SHOW_PROGRESS: bool = True
    ATHENA_PROGRESS_UPDATE_INTERVAL: float = 2.0  # Seconds between progress updates

    # Large query handling
    ATHENA_AUTO_CHUNK_LARGE_QUERIES: bool = True
    ATHENA_CHUNK_SIZE: int = 50  # Size for chunking large queries
    ATHENA_MAX_CONCURRENT_CHUNKS: int = 3  # Max concurrent chunk requests

    model_config = SettingsConfigDict(
        env_file=".env", env_prefix="", extra="ignore", env_file_encoding="utf-8"
    )


_settings_cache: Optional[_Settings] = None


def get_settings() -> _Settings:
    """
    Get the settings singleton.
    This function provides a cached instance of _Settings. The cache can be
    cleared using clear_settings_cache() when environment variables change.
    Returns:
        Cached instance of _Settings.
    """
    global _settings_cache
    if _settings_cache is None:
        _settings_cache = _Settings()
    return _settings_cache


def clear_settings_cache() -> None:
    """
    Clear the settings cache.
    This function should be called when environment variables change and you
    want the settings to be reloaded from the environment.
    """
    global _settings_cache
    _settings_cache = None


def reload_settings() -> _Settings:
    """
    Reload settings from environment variables.
    This function clears the cache and returns a fresh settings instance.
    Returns:
        Fresh instance of _Settings.
    """
    clear_settings_cache()
    return get_settings()



"""
Async HTTP client for the Athena API.

This module provides an asynchronous HTTP client for interacting with the Athena API.
It uses httpx for HTTP requests and provides automatic retry logic, rate limiting,
and error handling.
"""

import json
import logging
from typing import Any, Dict, Optional, Union, cast

try:
    import httpx
except ImportError as err:
    raise AttributeError(
        "httpx is required for the async client. Install with 'pip install httpx'"
    ) from err

try:
    import backoff
except ImportError as err:
    raise ImportError(
        "backoff is required for the async client. Install with 'pip install backoff'"
    ) from err

from .auth import build_headers
from .concept_explorer import create_concept_explorer
from .concept_set import ConceptSetGenerator
from .db.base import DatabaseConnector
from .exceptions import AthenaError, ClientError, NetworkError, ServerError
from .models import (
    ConceptDetails,
    ConceptRelationsGraph,
    ConceptRelationship,
    ConceptSearchResponse,
)
from .search_result import SearchResult
from .settings import get_settings

logger = logging.getLogger(__name__)


class AsyncHttpClient:
    """
    Asynchronous HTTP client for making requests to the Athena API.

    Features:
    - Automatic retry with exponential backoff
    - Custom timeout handling
    - Authentication header generation
    - Error handling and mapping to typed exceptions
    """

    def __init__(
        self,
        base_url: Optional[str] = None,
        token: Optional[str] = None,
        client_id: Optional[str] = None,
        private_key: Optional[str] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        backoff_factor: Optional[float] = None,
        db_connector: Optional[DatabaseConnector] = None,
    ) -> None:
        """
        Initialize the async HTTP client with configuration.

        Args:
            base_url: Base URL for the Athena API
            token: Bearer token for authentication
            client_id: Client ID for HMAC authentication
            private_key: Private key for HMAC signing
            timeout: HTTP timeout in seconds
            max_retries: Maximum number of retry attempts
            backoff_factor: Exponential backoff factor for retries
        """
        settings = get_settings()

        # Use provided values or fall back to settings
        self.base_url = base_url or settings.ATHENA_BASE_URL

        # Set up token and HMAC if provided
        if token is not None:
            settings.ATHENA_TOKEN = token
        if client_id is not None:
            settings.ATHENA_CLIENT_ID = client_id
        if private_key is not None:
            settings.ATHENA_PRIVATE_KEY = private_key

        self.timeout = timeout or settings.ATHENA_TIMEOUT_SECONDS
        self.max_retries = max_retries or settings.ATHENA_MAX_RETRIES
        self.backoff_factor = backoff_factor or settings.ATHENA_BACKOFF_FACTOR

        # Create httpx client
        self.client = httpx.AsyncClient(timeout=self.timeout)

        # Set up default headers
        self._setup_default_headers()

    def _setup_default_headers(self) -> None:
        """Set up default headers for all requests."""
        # Set default headers similar to the working client
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": "AthenaOHDSIAPIClient/1.0",
        }

        # Update client headers
        self.client.headers.update(default_headers)

        logger.debug("Default headers set: %s", default_headers)

    def _build_url(self, path: str) -> str:
        """
        Build full URL by joining base URL and path.

        Args:
            path: API endpoint path

        Returns:
            Full URL
        """
        # Handle paths that start with / to ensure they're appended correctly
        if path.startswith("/"):
            # Remove the leading / and join with base_url
            path = path[1:]

        # Ensure base_url doesn't end with / and path doesn't start with /
        if self.base_url.endswith("/"):
            base = self.base_url[:-1]
        else:
            base = self.base_url

        if path.startswith("/"):
            path = path[1:]

        full_url = f"{base}/{path}"

        logger.debug(
            f"Building URL: base_url='{self.base_url}', path='{path}' "
            f"-> full_url='{full_url}'"
        )
        return full_url

    async def _handle_response(self, response: httpx.Response) -> Dict[str, Any]:
        """
        Handle API response and raise appropriate exceptions.

        Args:
            response: HTTP response from httpx

        Returns:
            Parsed JSON response

        Raises:
            ClientError: For 4xx status codes
            ServerError: For 5xx status codes
            NetworkError: For connection errors
        """
        try:
            response.raise_for_status()
            return response.json()
        except httpx.HTTPStatusError as e:
            if 400 <= response.status_code < 500:
                raise ClientError(
                    f"Client error: {response.status_code} {response.reason_phrase}",
                    status_code=response.status_code,
                    response=response.text,
                ) from e
            elif response.status_code >= 500:
                raise ServerError(
                    f"Server error: {response.status_code} {response.reason_phrase}",
                    status_code=response.status_code,
                    response=response.text,
                ) from e
            else:
                raise
        except httpx.DecodingError as e:
            raise AthenaError(f"Invalid JSON response: {e}") from e

    @backoff.on_exception(
        backoff.expo,
        (httpx.TimeoutException, httpx.ConnectError),
        max_tries=3,
        factor=0.3,
    )
    async def request(
        self,
        method: str,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        raw_response: bool = False,
    ) -> Union[Dict[str, Any], httpx.Response]:
        """
        Make an HTTP request to the Athena API.

        Args:
            method: HTTP method (GET, POST, etc.)
            path: API endpoint path
            params: Query parameters
            data: Request body data
            raw_response: Whether to return the raw response object

        Returns:
            Parsed JSON response or raw Response object

        Raises:
            ClientError: For 4xx status codes
            ServerError: For 5xx status codes
            NetworkError: For connection errors
        """
        url = self._build_url(path)
        body_bytes = b""

        # Convert data to JSON bytes if provided
        if data is not None:
            body_bytes = json.dumps(data).encode("utf-8")

        # Build authentication headers
        headers = build_headers(method, url, body_bytes)

        # Add Content-Type header if sending data
        if data is not None:
            headers["Content-Type"] = "application/json"

        # Generate a correlation ID for logging
        correlation_id = f"req-{id(self)}-{id(path)}"
        logger.debug(f"[{correlation_id}] {method} {url}")

        try:
            response = await self.client.request(
                method=method,
                url=url,
                params=params,
                content=body_bytes if data is not None else None,
                headers=headers,
                timeout=self.timeout,
            )

            logger.debug(
                f"[{correlation_id}] {response.status_code} {response.reason_phrase}"
            )

            if raw_response:
                return response

            return await self._handle_response(response)

        except (httpx.TimeoutException, httpx.ConnectError) as e:
            logger.warning(f"[{correlation_id}] Network error: {e}")
            raise NetworkError(f"Network error: {e}") from e

    async def get(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        raw_response: bool = False,
    ) -> Union[Dict[str, Any], httpx.Response]:
        """
        Make a GET request to the Athena API.

        Args:
            path: API endpoint path
            params: Query parameters
            raw_response: Whether to return the raw response object

        Returns:
            Parsed JSON response or raw Response object
        """
        return await self.request("GET", path, params=params, raw_response=raw_response)

    async def post(
        self,
        path: str,
        data: Any = None,
        params: Optional[Dict[str, Any]] = None,
        raw_response: bool = False,
    ) -> Union[Dict[str, Any], httpx.Response]:
        """
        Make a POST request to the Athena API.

        Args:
            path: API endpoint path
            data: Request body data
            params: Query parameters
            raw_response: Whether to return the raw response object

        Returns:
            Parsed JSON response or raw Response object
        """
        return await self.request(
            "POST", path, data=data, params=params, raw_response=raw_response
        )


class AthenaAsyncClient:
    """
    Asynchronous client for the Athena API.

    This class provides asynchronous access to all Athena API endpoints
    with minimal abstraction, returning parsed Pydantic models.
    """

    def __init__(
        self,
        base_url: Optional[str] = None,
        token: Optional[str] = None,
        client_id: Optional[str] = None,
        private_key: Optional[str] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        backoff_factor: Optional[float] = None,
        db_connector: Optional[DatabaseConnector] = None,
    ) -> None:
        """
        Initialize the async Athena client with configuration.

        Args:
            base_url: Base URL for the Athena API
            token: Bearer token for authentication
            client_id: Client ID for HMAC authentication
            private_key: Private key for HMAC signing
            timeout: HTTP timeout in seconds
            max_retries: Maximum number of retry attempts
            backoff_factor: Exponential backoff factor for retries
            db_connector: Optional database connector for local OMOP validation
        """
        self.http = AsyncHttpClient(
            base_url=base_url,
            token=token,
            client_id=client_id,
            private_key=private_key,
            timeout=timeout,
            max_retries=max_retries,
            backoff_factor=backoff_factor,
        )
        self.db_connector = db_connector

    def set_database_connector(self, connector: DatabaseConnector) -> None:
        """Set the database connector for this client."""

        self.db_connector = connector

    async def search_concepts(
        self,
        query: str = "",
        exact: Optional[str] = None,
        fuzzy: bool = False,
        wildcard: Optional[str] = None,
        boosts: Optional[Dict[str, Any]] = None,
        debug: bool = False,
        page_size: int = 20,
        page: int = 0,
        domain: Optional[str] = None,
        vocabulary: Optional[str] = None,
        standard_concept: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Search for concepts in the Athena vocabulary.

        Args:
            query: The search query string
            exact: Exact match phrase
            fuzzy: Whether to enable fuzzy matching
            wildcard: Wildcard pattern
            boosts: Dictionary of field boosts
            debug: Enable debug mode
            page_size: Number of results per page
            page: Page number (0-indexed)
            domain: Filter by domain
            vocabulary: Filter by vocabulary
            standard_concept: Filter by standard concept status

        Returns:
            Raw API response data
        """
        # Convert page to start parameter that the API expects
        start = page * page_size

        params: Dict[str, Any] = {"pageSize": page_size, "start": start}

        # Add query if provided
        if query:
            params["query"] = query

        # Add filters if provided
        if exact:
            params["exact"] = exact
        if fuzzy:
            params["fuzzy"] = str(fuzzy).lower()
        if wildcard:
            params["wildcard"] = wildcard
        if domain:
            params["domain"] = domain
        if vocabulary:
            params["vocabulary"] = vocabulary
        if standard_concept:
            params["standardConcept"] = standard_concept

        # If boosts provided, use debug endpoint and include boosts in request
        if boosts or debug:
            response = await self.http.post(
                "/concepts",
                data={"boosts": boosts} if boosts else {},
                params=params,
            )
            return cast(Dict[str, Any], response)

        # Otherwise use standard GET endpoint
        response = await self.http.get("/concepts", params=params)
        return cast(Dict[str, Any], response)

    async def search(
        self,
        query: str = "",
        size: int = 20,
        page: int = 0,
        **kwargs: Any,
    ) -> SearchResult:
        """
        Search for concepts with a SearchResult wrapper.

        Args:
            query: The search query string
            size: Number of results per page
            page: Page number (0-indexed)
            **kwargs: Additional search parameters

        Returns:
            SearchResult object with convenient access methods
        """
        # Convert size to pageSize for the API
        search_kwargs: Dict[str, Any] = dict(kwargs)
        search_kwargs["page_size"] = size
        search_kwargs["page"] = page

        response_data = await self.search_concepts(query=query, **search_kwargs)
        response = ConceptSearchResponse.model_validate(response_data)
        return SearchResult(response, self)

    async def details(self, concept_id: int) -> ConceptDetails:
        """
        Get detailed information for a specific concept.

        This is an alias for get_concept_details for compatibility.

        Args:
            concept_id: The concept ID to get details for

        Returns:
            ConceptDetails object
        """
        return await self.get_concept_details(concept_id)

    async def relationships(
        self,
        concept_id: int,
        only_standard: bool = False,
        relationship_id: Optional[str] = None,
    ) -> ConceptRelationship:
        """
        Get relationships for a specific concept.

        This is an alias for get_concept_relationships for compatibility.

        Args:
            concept_id: The concept ID to get relationships for
            only_standard: Only include standard concepts
            relationship_id: Filter by relationship type

        Returns:
            ConceptRelationship object
        """
        return await self.get_concept_relationships(
            concept_id, relationship_id=relationship_id, only_standard=only_standard
        )

    async def get_concept_details(self, concept_id: int) -> ConceptDetails:
        """
        Get detailed information for a specific concept.

        Args:
            concept_id: The concept ID to get details for

        Returns:
            ConceptDetails object
        """
        response = await self.http.get(f"/concepts/{concept_id}")
        data = cast(Dict[str, Any], response)
        return ConceptDetails.model_validate(data)

    async def get_concept_relationships(
        self,
        concept_id: int,
        relationship_id: Optional[str] = None,
        only_standard: bool = False,
    ) -> ConceptRelationship:
        """
        Get relationships for a specific concept.

        Args:
            concept_id: The concept ID to get relationships for
            relationship_id: Filter by relationship type
            only_standard: Only include standard concepts

        Returns:
            ConceptRelationship object
        """
        params: Dict[str, Any] = {}

        if relationship_id:
            params["relationshipId"] = relationship_id
        if only_standard:
            params["standardConcepts"] = "true"

        response = await self.http.get(
            f"/concepts/{concept_id}/relationships", params=params
        )
        data = cast(Dict[str, Any], response)
        return ConceptRelationship.model_validate(data)

    async def get_concept_graph(
        self,
        concept_id: int,
        depth: int = 10,
        zoom_level: int = 4,
    ) -> ConceptRelationsGraph:
        """
        Get relationship graph for a specific concept.

        Args:
            concept_id: The concept ID to get graph for
            depth: Maximum depth of relationships to traverse
            zoom_level: Zoom level for the graph

        Returns:
            ConceptRelationsGraph object
        """
        params = {"depth": depth, "zoomLevel": zoom_level}
        response = await self.http.get(
            f"/concepts/{concept_id}/relations", params=params
        )
        data = cast(Dict[str, Any], response)
        return ConceptRelationsGraph.model_validate(data)

    async def generate_concept_set(self, query: str, **kwargs: Any) -> Dict[str, Any]:
        """Generate a validated concept set from a search query."""

        if not self.db_connector:
            raise RuntimeError("A database connector has not been configured.")

        explorer = create_concept_explorer(self)
        generator = ConceptSetGenerator(explorer=explorer, db=self.db_connector)

        return await generator.create_from_query(query, **kwargs)



"""Tests for the utils module."""

import logging
from unittest.mock import Mock, patch

from athena_client.utils import configure_logging


class TestUtils:
    """Test cases for the utils module."""

    def test_configure_logging_default_level(self):
        """Test logging configuration with default level."""
        with patch("athena_client.utils.logging") as mock_logging:
            mock_logger = Mock()
            mock_logging.getLogger.return_value = mock_logger
            mock_logging.INFO = logging.INFO

            configure_logging()

            mock_logging.getLogger.assert_called_with("athena_client")
            mock_logger.setLevel.assert_called_with(logging.INFO)

    def test_configure_logging_custom_level(self):
        """Test logging configuration with custom level."""
        with patch("athena_client.utils.logging") as mock_logging:
            mock_logger = Mock()
            mock_logging.getLogger.return_value = mock_logger
            mock_logging.DEBUG = logging.DEBUG

            configure_logging(logging.DEBUG)

            mock_logger.setLevel.assert_called_with(logging.DEBUG)

    def test_configure_logging_with_existing_handlers(self):
        """Test logging configuration when handlers already exist."""
        with patch("athena_client.utils.logging") as mock_logging:
            mock_logger = Mock()
            mock_logger.handlers = [Mock()]  # Existing handler
            mock_logging.getLogger.return_value = mock_logger
            mock_logging.INFO = logging.INFO
            mock_logging.StreamHandler = Mock()
            mock_logging.Formatter = Mock()

            configure_logging()

            # Should not add new handler when handlers already exist
            mock_logging.StreamHandler.assert_not_called()

    def test_configure_logging_without_handlers(self):
        """Test logging configuration when no handlers exist."""
        with patch("athena_client.utils.logging") as mock_logging:
            mock_logger = Mock()
            mock_logger.handlers = []  # No existing handlers
            mock_logging.getLogger.return_value = mock_logger
            mock_logging.INFO = logging.INFO

            mock_handler = Mock()
            mock_logging.StreamHandler.return_value = mock_handler

            mock_formatter = Mock()
            mock_logging.Formatter.return_value = mock_formatter

            configure_logging()

            # Should create new handler and formatter
            mock_logging.StreamHandler.assert_called_once()
            mock_logging.Formatter.assert_called_once()
            mock_handler.setLevel.assert_called_with(logging.INFO)
            mock_handler.setFormatter.assert_called_with(mock_formatter)
            mock_logger.addHandler.assert_called_with(mock_handler)

    def test_configure_logging_none_level(self):
        """Test logging configuration with None level."""
        with patch("athena_client.utils.logging") as mock_logging:
            mock_logger = Mock()
            mock_logging.getLogger.return_value = mock_logger
            mock_logging.INFO = logging.INFO

            configure_logging(None)

            # Should use default INFO level
            mock_logger.setLevel.assert_called_with(logging.INFO)

    def test_configure_logging_formatter_format(self):
        """Test that the formatter is created with the correct format."""
        with patch("athena_client.utils.logging") as mock_logging:
            mock_logger = Mock()
            mock_logger.handlers = []
            mock_logging.getLogger.return_value = mock_logger
            mock_logging.INFO = logging.INFO

            mock_handler = Mock()
            mock_logging.StreamHandler.return_value = mock_handler

            mock_formatter = Mock()
            mock_logging.Formatter.return_value = mock_formatter

            configure_logging()

            # Should create formatter with the correct format string
            expected_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            mock_logging.Formatter.assert_called_with(expected_format)

    def test_configure_logging_handler_level(self):
        """Test that the handler level is set correctly."""
        with patch("athena_client.utils.logging") as mock_logging:
            mock_logger = Mock()
            mock_logger.handlers = []
            mock_logging.getLogger.return_value = mock_logger
            mock_logging.INFO = logging.INFO

            mock_handler = Mock()
            mock_logging.StreamHandler.return_value = mock_handler

            mock_formatter = Mock()
            mock_logging.Formatter.return_value = mock_formatter

            configure_logging(logging.DEBUG)

            # Handler level should match the configured level
            mock_handler.setLevel.assert_called_with(logging.DEBUG)

    def test_configure_logging_logger_name(self):
        """Test that the correct logger name is used."""
        with patch("athena_client.utils.logging") as mock_logging:
            mock_logger = Mock()
            mock_logging.getLogger.return_value = mock_logger
            mock_logging.INFO = logging.INFO

            configure_logging()

            # Should get logger with the correct name
            mock_logging.getLogger.assert_called_with("athena_client")

    def test_configure_logging_multiple_calls(self):
        """Test that multiple calls to configure_logging work correctly."""
        with patch("athena_client.utils.logging") as mock_logging:
            mock_logger = Mock()
            mock_logger.handlers = []
            mock_logging.getLogger.return_value = mock_logger
            mock_logging.INFO = logging.INFO
            mock_logging.DEBUG = logging.DEBUG

            mock_handler = Mock()
            mock_logging.StreamHandler.return_value = mock_handler

            mock_formatter = Mock()
            mock_logging.Formatter.return_value = mock_formatter

            # First call
            configure_logging(logging.INFO)

            # Second call with different level
            configure_logging(logging.DEBUG)

            # Should set level to DEBUG on second call
            assert mock_logger.setLevel.call_count == 2
            mock_logger.setLevel.assert_called_with(logging.DEBUG)

    def test_configure_logging_integration(self):
        """Test logging configuration integration with real logging."""
        # Test with real logging module
        original_handlers = logging.getLogger("athena_client").handlers[:]

        try:
            # Configure logging
            configure_logging(logging.DEBUG)

            # Get the logger
            logger = logging.getLogger("athena_client")

            # Check that level is set correctly
            assert logger.level == logging.DEBUG

            # Check that handlers were added
            assert len(logger.handlers) > 0

            # Check that handler level is set correctly
            for handler in logger.handlers:
                if isinstance(handler, logging.StreamHandler):
                    assert handler.level == logging.DEBUG
                    assert handler.formatter is not None

        finally:
            # Clean up - remove handlers we added
            logger = logging.getLogger("athena_client")
            for handler in logger.handlers[:]:
                if handler not in original_handlers:
                    logger.removeHandler(handler)



"""
Exception classes for the Athena client.

This module defines a hierarchy of exceptions that can be raised by the Athena client.
"""

from typing import List, Optional


class AthenaError(Exception):
    """Base class for all Athena client exceptions."""

    def __init__(
        self,
        message: str,
        error_code: Optional[str] = None,
        troubleshooting: Optional[str] = None,
    ) -> None:
        """
        Initialize the exception.

        Args:
            message: Error message
            error_code: Optional error code for programmatic handling
            troubleshooting: Optional troubleshooting suggestions
        """
        super().__init__(message)
        self.message = message
        self.error_code = error_code
        self.troubleshooting = troubleshooting

    def __str__(self) -> str:
        """Return a user-friendly error message."""
        msg = self.message
        if self.troubleshooting:
            msg += f"\n\n💡 Troubleshooting: {self.troubleshooting}"
        return msg


class NetworkError(AthenaError):
    """
    Raised for network-related errors (DNS, TLS, socket, or timeout).
    """

    def __init__(self, message: str, url: Optional[str] = None) -> None:
        """
        Initialize the network error.

        Args:
            message: Error message
            url: URL that failed to connect
        """
        troubleshooting = (
            "• Check your internet connection\n"
            "• Verify the API endpoint URL is correct\n"
            "• Try again in a few moments\n"
            "• Contact your network administrator if the problem persists"
        )
        super().__init__(
            message, error_code="NETWORK_ERROR", troubleshooting=troubleshooting
        )
        self.url = url


class TimeoutError(NetworkError):
    """
    Raised when a request times out.
    """

    def __init__(
        self, message: str, url: Optional[str] = None, timeout: Optional[float] = None
    ) -> None:
        """
        Initialize the timeout error.

        Args:
            message: Error message
            url: URL that timed out
            timeout: Timeout value that was exceeded
        """
        troubleshooting = (
            "• The server is taking too long to respond\n"
            "• Try increasing the timeout value\n"
            "• Check if the API server is experiencing high load\n"
            "• Try again in a few moments"
        )
        super().__init__(message, url)
        self.timeout = timeout
        self.troubleshooting = troubleshooting


class ServerError(AthenaError):
    """
    Raised when the server returns a 5xx status code.
    """

    def __init__(
        self,
        message: str,
        status_code: int,
        response: Optional[str] = None,
        url: Optional[str] = None,
    ) -> None:
        """
        Initialize the server error.

        Args:
            message: Error message
            status_code: HTTP status code
            response: Raw response body
            url: URL that caused the error
        """
        troubleshooting = (
            "• The API server is experiencing issues\n"
            "• Try again in a few moments\n"
            "• Check the API status page for known issues\n"
            "• Contact the API administrators if the problem persists"
        )
        super().__init__(
            message, error_code="SERVER_ERROR", troubleshooting=troubleshooting
        )
        self.status_code = status_code
        self.response = response
        self.url = url


class ClientError(AthenaError):
    """
    Raised when the server returns a 4xx status code.
    """

    def __init__(
        self,
        message: str,
        status_code: int,
        response: Optional[str] = None,
        url: Optional[str] = None,
    ) -> None:
        """
        Initialize the client error.

        Args:
            message: Error message
            status_code: HTTP status code
            response: Raw response body
            url: URL that caused the error
        """
        # Customize troubleshooting based on status code
        if status_code == 400:
            troubleshooting = (
                "• Check your request parameters\n"
                "• Verify the data format is correct\n"
                "• Review the API documentation for the correct format"
            )
        elif status_code == 401:
            troubleshooting = (
                "• Check your authentication credentials\n"
                "• Verify your API token is valid and not expired\n"
                "• Ensure you have the necessary permissions"
            )
        elif status_code == 403:
            troubleshooting = (
                "• You don't have permission to access this resource\n"
                "• Check your API token permissions\n"
                "• Contact the API administrators for access"
            )
        elif status_code == 404:
            troubleshooting = (
                "• The requested resource was not found\n"
                "• Check the URL path is correct\n"
                "• Verify the resource ID exists\n"
                "• Review the API documentation for available endpoints"
            )
        elif status_code == 429:
            troubleshooting = (
                "• You've exceeded the rate limit\n"
                "• Wait before making more requests\n"
                "• Consider implementing request throttling\n"
                "• Check the API documentation for rate limits"
            )
        else:
            troubleshooting = (
                "• Check your request parameters\n"
                "• Review the API documentation\n"
                "• Verify your authentication credentials"
            )

        super().__init__(
            message, error_code="CLIENT_ERROR", troubleshooting=troubleshooting
        )
        self.status_code = status_code
        self.response = response
        self.url = url


class ValidationError(AthenaError):
    """
    Raised when response validation fails.
    """

    def __init__(self, message: str, validation_details: Optional[str] = None) -> None:
        """
        Initialize the validation error.

        Args:
            message: Error message
            validation_details: Detailed validation error information
        """
        troubleshooting = (
            "• The API response format has changed\n"
            "• This might be a temporary API issue\n"
            "• Try again in a few moments\n"
            "• Contact the API administrators if the problem persists"
        )
        super().__init__(
            message, error_code="VALIDATION_ERROR", troubleshooting=troubleshooting
        )
        self.validation_details = validation_details


class AuthenticationError(ClientError):
    """
    Raised for authentication-related errors.
    """

    def __init__(
        self,
        message: str,
        status_code: int = 401,
        response: Optional[str] = None,
        url: Optional[str] = None,
    ) -> None:
        """
        Initialize the authentication error.

        Args:
            message: Error message
            status_code: HTTP status code (default 401)
            response: Raw response body
            url: URL that caused the error
        """
        troubleshooting = (
            "• Check your API token is valid and not expired\n"
            "• Verify your authentication credentials\n"
            "• Ensure you have the necessary permissions\n"
            "• Contact the API administrators for access"
        )
        super().__init__(message, status_code, response, url)
        self.error_code = "AUTHENTICATION_ERROR"
        self.troubleshooting = troubleshooting


class RateLimitError(ClientError):
    """
    Raised when rate limits are exceeded.
    """

    def __init__(
        self,
        message: str,
        status_code: int = 429,
        response: Optional[str] = None,
        url: Optional[str] = None,
    ) -> None:
        """
        Initialize the rate limit error.

        Args:
            message: Error message
            status_code: HTTP status code (default 429)
            response: Raw response body
            url: URL that caused the error
        """
        troubleshooting = (
            "• You've exceeded the API rate limit\n"
            "• Wait before making more requests\n"
            "• Consider implementing request throttling\n"
            "• Check the API documentation for rate limits"
        )
        super().__init__(message, status_code, response, url)
        self.error_code = "RATE_LIMIT_ERROR"
        self.troubleshooting = troubleshooting


class APIError(AthenaError):
    """
    Raised when the API returns an error response.
    """

    def __init__(
        self,
        message: str,
        api_error_code: Optional[str] = None,
        api_message: Optional[str] = None,
    ) -> None:
        """
        Initialize the API error.

        Args:
            message: Error message
            api_error_code: Error code from the API
            api_message: Error message from the API
        """
        troubleshooting = (
            "• The API returned an error response\n"
            "• Check the API error details above\n"
            "• Verify your request parameters\n"
            "• Try again with different parameters if applicable"
        )
        super().__init__(
            message, error_code="API_ERROR", troubleshooting=troubleshooting
        )
        self.api_error_code = api_error_code
        self.api_message = api_message


class RetryFailedError(AthenaError):
    """
    Raised when all retry attempts have been exhausted.
    """

    def __init__(
        self,
        message: str,
        max_attempts: int,
        last_error: Exception,
        retry_history: List[Exception],
        error_code: Optional[str] = None,
    ) -> None:
        """
        Initialize the retry failed error.

        Args:
            message: Error message
            max_attempts: Maximum number of retry attempts
            last_error: The last error that caused the final failure
            retry_history: List of errors from each retry attempt
            error_code: Optional error code
        """
        troubleshooting = (
            "• All retry attempts have been exhausted\n"
            "• Check your internet connection\n"
            "• Verify the API server is accessible\n"
            "• Try again in a few moments\n"
            "• Consider increasing max_retries if this is a temporary issue"
        )
        super().__init__(
            message,
            error_code=error_code or "RETRY_FAILED",
            troubleshooting=troubleshooting,
        )
        self.max_attempts = max_attempts
        self.last_error = last_error
        self.retry_history = retry_history

    def __str__(self) -> str:
        """Return a detailed error message with retry information."""
        msg = self.message
        msg += "\n\n📊 Retry Information:"
        msg += f"\n• Maximum attempts: {self.max_attempts}"
        msg += f"\n• Attempts made: {len(self.retry_history) + 1}"
        msg += f"\n• Last error: {type(self.last_error).__name__}: {self.last_error}"

        if self.retry_history:
            msg += "\n\n🔄 Retry History:"
            for i, error in enumerate(self.retry_history, 1):
                msg += f"\n• Attempt {i}: {type(error).__name__}: {error}"

        if self.troubleshooting:
            msg += f"\n\n💡 Troubleshooting: {self.troubleshooting}"

        return msg



"""
Tests for settings module to ensure proper handling of environment variables.
"""

import os
from unittest.mock import patch

from athena_client.settings import (
    _Settings,
    clear_settings_cache,
    get_settings,
    reload_settings,
)


class TestSettingsEnvironmentVariables:
    """Test settings behavior with various environment variable configurations."""

    def setup_method(self):
        """Clear settings cache before each test to ensure clean state."""
        clear_settings_cache()

    def test_settings_initialization_without_external_env_vars(self):
        """Test that settings initialize correctly without unrelated env vars."""
        # Clear any existing environment variables that might interfere
        with patch.dict(os.environ, {}, clear=True):
            settings = get_settings()
            assert settings.ATHENA_BASE_URL == "https://athena.ohdsi.org/api/v1"
            assert settings.ATHENA_TIMEOUT_SECONDS == 30

    def test_settings_with_external_google_credentials(self):
        """Test that settings work when GOOGLE_APPLICATION_CREDENTIALS is set."""
        test_env = {
            "GOOGLE_APPLICATION_CREDENTIALS": "/path/to/service-account.json",
            "ATHENA_BASE_URL": "https://custom.athena.org/api/v1",
        }

        with patch.dict(os.environ, test_env, clear=True):
            settings = get_settings()
            # Should not fail and should use the custom ATHENA_BASE_URL
            assert settings.ATHENA_BASE_URL == "https://custom.athena.org/api/v1"
            assert settings.ATHENA_TIMEOUT_SECONDS == 30

    def test_settings_with_external_openai_key(self):
        """Test that settings work when OPENAI_API_KEY is set."""
        test_env = {
            "OPENAI_API_KEY": "sk-proj-vO1urVisRlM1Upkk...AtK0BzkoftVcCEhpvShK4wA",
            "ATHENA_TOKEN": "test-token",
        }

        with patch.dict(os.environ, test_env, clear=True):
            settings = get_settings()
            # Should not fail and should use the ATHENA_TOKEN
            assert settings.ATHENA_TOKEN == "test-token"
            assert settings.ATHENA_BASE_URL == "https://athena.ohdsi.org/api/v1"

    def test_settings_with_multiple_external_vars(self):
        """Test that settings work with multiple external environment variables."""
        test_env = {
            "GOOGLE_APPLICATION_CREDENTIALS": "/path/to/service-account.json",
            "OPENAI_API_KEY": "sk-proj-vO1urVisRlM1Upkk...AtK0BzkoftVcCEhpvShK4wA",
            "AWS_ACCESS_KEY_ID": "AKIAIOSFODNN7EXAMPLE",
            "AWS_SECRET_ACCESS_KEY": "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY",
            "ATHENA_TOKEN": "test-token",
            "ATHENA_TIMEOUT_SECONDS": "60",
        }

        with patch.dict(os.environ, test_env, clear=True):
            settings = get_settings()
            # Should not fail and should use the athena-specific settings
            assert settings.ATHENA_TOKEN == "test-token"
            assert settings.ATHENA_TIMEOUT_SECONDS == 60
            assert settings.ATHENA_BASE_URL == "https://athena.ohdsi.org/api/v1"

    def test_settings_extra_fields_ignored(self):
        """Test that extra fields are properly ignored."""
        # Create a settings instance with extra fields
        test_env = {
            "EXTRA_FIELD_1": "value1",
            "EXTRA_FIELD_2": "value2",
            "ATHENA_BASE_URL": "https://test.athena.org/api/v1",
        }

        with patch.dict(os.environ, test_env, clear=True):
            settings = get_settings()
            # Should not fail and should only use ATHENA_ prefixed fields
            assert settings.ATHENA_BASE_URL == "https://test.athena.org/api/v1"

            # Verify that extra fields are not accessible
            assert not hasattr(settings, "EXTRA_FIELD_1")
            assert not hasattr(settings, "EXTRA_FIELD_2")

    def test_settings_cache_behavior(self):
        """Test that settings caching works correctly with environment changes."""
        # First call with default environment
        with patch.dict(os.environ, {}, clear=True):
            settings1 = get_settings()
            assert settings1.ATHENA_BASE_URL == "https://athena.ohdsi.org/api/v1"

        # Clear cache and call with custom environment
        clear_settings_cache()

        with patch.dict(
            os.environ,
            {"ATHENA_BASE_URL": "https://custom.athena.org/api/v1"},
            clear=True,
        ):
            settings2 = get_settings()
            assert settings2.ATHENA_BASE_URL == "https://custom.athena.org/api/v1"

    def test_settings_model_config(self):
        """Test that the model configuration is set correctly."""
        # Verify that the model config ignores extra fields
        assert _Settings.model_config.get("extra") == "ignore"
        assert _Settings.model_config.get("env_file") == ".env"
        assert _Settings.model_config.get("env_prefix") == ""

    def test_settings_all_defined_fields(self):
        """Test that all defined settings fields are accessible."""
        settings = get_settings()

        # Test all the defined fields exist and have expected types
        assert isinstance(settings.ATHENA_BASE_URL, str)
        assert isinstance(settings.ATHENA_TOKEN, (str, type(None)))
        assert isinstance(settings.ATHENA_CLIENT_ID, (str, type(None)))
        assert isinstance(settings.ATHENA_PRIVATE_KEY, (str, type(None)))
        assert isinstance(settings.ATHENA_TIMEOUT_SECONDS, int)
        assert isinstance(settings.ATHENA_SEARCH_TIMEOUT_SECONDS, int)
        assert isinstance(settings.ATHENA_GRAPH_TIMEOUT_SECONDS, int)
        assert isinstance(settings.ATHENA_RELATIONSHIPS_TIMEOUT_SECONDS, int)
        assert isinstance(settings.ATHENA_MAX_RETRIES, int)
        assert isinstance(settings.ATHENA_BACKOFF_FACTOR, float)
        assert isinstance(settings.ATHENA_DEFAULT_PAGE_SIZE, int)
        assert isinstance(settings.ATHENA_MAX_PAGE_SIZE, int)
        assert isinstance(settings.ATHENA_LARGE_QUERY_THRESHOLD, int)
        assert isinstance(settings.ATHENA_SHOW_PROGRESS, bool)
        assert isinstance(settings.ATHENA_PROGRESS_UPDATE_INTERVAL, float)
        assert isinstance(settings.ATHENA_AUTO_CHUNK_LARGE_QUERIES, bool)
        assert isinstance(settings.ATHENA_CHUNK_SIZE, int)
        assert isinstance(settings.ATHENA_MAX_CONCURRENT_CHUNKS, int)

    def test_clear_settings_cache(self):
        """Test that clear_settings_cache works correctly."""
        # Get initial settings
        settings1 = get_settings()

        # Clear cache
        clear_settings_cache()

        # Get settings again - should be a new instance
        settings2 = get_settings()

        # They should be different instances
        assert settings1 is not settings2

    def test_reload_settings(self):
        """Test that reload_settings works correctly."""
        # Get initial settings
        settings1 = get_settings()

        # Reload settings
        settings2 = reload_settings()

        # They should be different instances
        assert settings1 is not settings2

    def test_external_env_vars_do_not_cause_errors(self):
        """Test that external environment variables don't cause validation errors."""
        test_env = {
            "GOOGLE_APPLICATION_CREDENTIALS": "/path/to/service-account.json",
            "OPENAI_API_KEY": "sk-proj-vO1urVisRlM1Upkk...AtK0BzkoftVcCEhpvShK4wA",
            "AWS_ACCESS_KEY_ID": "AKIAIOSFODNN7EXAMPLE",
            "AWS_SECRET_ACCESS_KEY": "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY",
            "DATABASE_URL": "postgresql://user:pass@localhost/db",
            "REDIS_URL": "redis://localhost:6379",
            "ATHENA_TOKEN": "test-token",
        }

        with patch.dict(os.environ, test_env, clear=True):
            # This should not raise any validation errors
            settings = get_settings()
            assert settings.ATHENA_TOKEN == "test-token"
            assert settings.ATHENA_BASE_URL == "https://athena.ohdsi.org/api/v1"

    def test_only_athena_prefixed_vars_are_used(self):
        """Test that only ATHENA_ prefixed environment variables are used."""
        test_env = {
            "ATHENA_TOKEN": "athena-token",
            "ATHENA_BASE_URL": "https://custom.athena.org/api/v1",
            "TOKEN": "other-token",  # This should be ignored
            "BASE_URL": "https://other.org/api/v1",  # This should be ignored
        }

        with patch.dict(os.environ, test_env, clear=True):
            settings = get_settings()
            assert settings.ATHENA_TOKEN == "athena-token"
            assert settings.ATHENA_BASE_URL == "https://custom.athena.org/api/v1"

            # Verify that non-ATHENA_ prefixed variables are not accessible
            assert not hasattr(settings, "TOKEN")
            assert not hasattr(settings, "BASE_URL")



from typing import Dict, List

from sqlalchemy import bindparam, create_engine, text
from sqlalchemy.engine import Engine


class SQLAlchemyConnector:
    """Database connector using SQLAlchemy Core."""

    def __init__(self, engine: Engine) -> None:
        self._engine = engine

    def validate_concepts(self, concept_ids: List[int]) -> List[int]:
        if not concept_ids:
            return []

        stmt = text(
            """
                SELECT concept_id
                FROM concept
                WHERE concept_id IN :ids AND standard_concept = 'S'
                """
        ).bindparams(bindparam("ids", expanding=True))

        with self._engine.connect() as connection:
            result = connection.execute(stmt, {"ids": list(concept_ids)})
            validated_ids = [row[0] for row in result]

        return validated_ids

    def get_descendants(self, concept_ids: List[int]) -> List[int]:
        """Retrieve descendant concept IDs for the given ancestors."""
        if not concept_ids:
            return []

        stmt = text(
            """
                SELECT descendant_concept_id
                FROM concept_ancestor
                WHERE ancestor_concept_id IN :ids
                """
        ).bindparams(bindparam("ids", expanding=True))

        with self._engine.connect() as connection:
            result = connection.execute(stmt, {"ids": list(concept_ids)})
            descendant_ids = [row[0] for row in result]

        return list(set(descendant_ids) - set(concept_ids))

    def get_standard_mapping(
        self, non_standard_concept_ids: List[int]
    ) -> Dict[int, int]:
        """Find standard mappings for the given non-standard concept IDs."""
        if not non_standard_concept_ids:
            return {}

        stmt = text(
            """
            SELECT cr.concept_id_1, cr.concept_id_2, c2.standard_concept
            FROM concept_relationship cr
            JOIN concept c2 ON cr.concept_id_2 = c2.concept_id
            WHERE cr.concept_id_1 IN :ids
              AND cr.relationship_id = 'Maps to'
              AND cr.invalid_reason IS NULL
            """
        ).bindparams(bindparam("ids", expanding=True))

        with self._engine.connect() as connection:
            result = connection.execute(stmt, {"ids": list(non_standard_concept_ids)})
            mapping: Dict[int, int] = {}
            for row in result:
                concept_id_1, concept_id_2, standard_flag = row
                if standard_flag == "S" and concept_id_1 not in mapping:
                    mapping[concept_id_1] = concept_id_2

        return mapping

    @staticmethod
    def from_connection_string(connection_string: str) -> "SQLAlchemyConnector":
        engine = create_engine(connection_string)
        return SQLAlchemyConnector(engine)



"""
Tests for the Query DSL.
"""

from athena_client.query import Q


def test_term_query():
    """Test basic term query construction."""
    q = Q.term("diabetes")
    boosts = q.to_boosts()

    assert boosts["filter"]["query_string"]["query"] == "diabetes"
    assert "name^10" in boosts["filter"]["query_string"]["fields"]
    assert "synonyms^5" in boosts["filter"]["query_string"]["fields"]


def test_phrase_query():
    """Test phrase query construction."""
    q = Q.phrase("heart attack")
    boosts = q.to_boosts()

    assert boosts["filter"]["query_string"]["query"] == '"heart attack"'


def test_exact_query():
    """Test exact query construction."""
    q = Q.exact('"Myocardial Infarction"')
    boosts = q.to_boosts()

    assert boosts["filter"]["query_string"]["query"] == '"Myocardial Infarction"'
    assert "concept_code^20" in boosts["filter"]["query_string"]["fields"]
    assert "name.exact^15" in boosts["filter"]["query_string"]["fields"]


def test_wildcard_query():
    """Test wildcard query construction."""
    q = Q.wildcard("aspir*")
    boosts = q.to_boosts()

    assert boosts["filter"]["query_string"]["query"] == "aspir*"


def test_fuzzy_modifier():
    """Test fuzzy modifier."""
    q = Q.term("aspirin").fuzzy(0.8)
    boosts = q.to_boosts()

    assert boosts["filter"]["query_string"]["fuzziness"] == 0.8


def test_and_operator():
    """Test AND operator."""
    q = Q.term("diabetes") & Q.term("type 2")
    boosts = q.to_boosts()

    assert "must" in boosts["filter"]["bool"]
    assert len(boosts["filter"]["bool"]["must"]) == 2


def test_or_operator():
    """Test OR operator."""
    q = Q.term("heart") | Q.term("cardiac")
    boosts = q.to_boosts()

    assert "should" in boosts["filter"]["bool"]
    assert len(boosts["filter"]["bool"]["should"]) == 2


def test_not_operator():
    """Test NOT operator."""
    q = -Q.term("chronic")
    boosts = q.to_boosts()

    assert "must_not" in boosts["filter"]["bool"]
    assert len(boosts["filter"]["bool"]["must_not"]) == 1


def test_complex_query():
    """Test complex query with multiple operators."""
    q = (Q.term("diabetes") & Q.term("type 2")) | Q.exact('"diabetic nephropathy"')
    boosts = q.to_boosts()

    assert "should" in boosts["filter"]["bool"]



"""Tests for the authentication module."""

from unittest.mock import Mock, patch

from athena_client.auth import build_headers


class TestAuth:
    """Test cases for the authentication module."""

    def test_build_headers_no_auth(self):
        """Test building headers without authentication."""
        with patch("athena_client.auth.get_settings") as mock_get_settings:
            mock_settings = Mock()
            mock_settings.ATHENA_TOKEN = None
            mock_settings.ATHENA_CLIENT_ID = None
            mock_settings.ATHENA_PRIVATE_KEY = None
            mock_get_settings.return_value = mock_settings

            headers = build_headers("GET", "https://api.example.com/test", b"")

            assert headers == {}

    def test_build_headers_with_token(self):
        """Test building headers with Bearer token authentication."""
        with patch("athena_client.auth.get_settings") as mock_get_settings:
            mock_settings = Mock()
            mock_settings.ATHENA_TOKEN = "test-token"
            mock_settings.ATHENA_CLIENT_ID = "test-client"
            mock_settings.ATHENA_PRIVATE_KEY = None
            mock_get_settings.return_value = mock_settings

            headers = build_headers("GET", "https://api.example.com/test", b"")

            assert headers["X-Athena-Auth"] == "Bearer test-token"
            assert headers["X-Athena-Client-Id"] == "test-client"

    def test_build_headers_with_token_no_client_id(self):
        """Test building headers with token but no client ID."""
        with patch("athena_client.auth.get_settings") as mock_get_settings:
            mock_settings = Mock()
            mock_settings.ATHENA_TOKEN = "test-token"
            mock_settings.ATHENA_CLIENT_ID = None
            mock_settings.ATHENA_PRIVATE_KEY = None
            mock_get_settings.return_value = mock_settings

            headers = build_headers("GET", "https://api.example.com/test", b"")

            assert headers["X-Athena-Auth"] == "Bearer test-token"
            assert headers["X-Athena-Client-Id"] == "athena-client"

    def test_build_headers_with_hmac(self):
        """Test building headers with HMAC authentication."""
        with patch("athena_client.auth.get_settings") as mock_get_settings:
            mock_settings = Mock()
            mock_settings.ATHENA_TOKEN = None
            mock_settings.ATHENA_CLIENT_ID = None
            mock_settings.ATHENA_PRIVATE_KEY = (
                "-----BEGIN PRIVATE KEY-----\ntest-key\n-----END PRIVATE KEY-----"
            )
            mock_get_settings.return_value = mock_settings

            # Mock the imports that happen inside build_headers
            mock_datetime = Mock()
            mock_datetime.utcnow.return_value.strftime.return_value = (
                "2023-01-01T00:00:00Z"
            )

            mock_b64encode = Mock()
            mock_b64encode.return_value.decode.return_value = "test-signature"

            mock_serialization = Mock()
            mock_key = Mock()
            mock_serialization.load_pem_private_key.return_value = mock_key
            mock_key.sign.return_value = b"test-signature"

            mock_hashes = Mock()
            mock_hashes.SHA256.return_value = "sha256"

            with (
                patch("athena_client.auth.datetime", mock_datetime),
                patch("athena_client.auth.b64encode", mock_b64encode),
            ):
                headers = build_headers(
                    "GET",
                    "https://api.example.com/test",
                    b"",
                    serialization_module=mock_serialization,
                    hashes_module=mock_hashes,
                )
                assert "X-Athena-Nonce" in headers
                assert "X-Athena-Hmac" in headers
                assert headers["X-Athena-Nonce"] == "2023-01-01T00:00:00Z"
                assert headers["X-Athena-Hmac"] == "test-signature"

    def test_build_headers_with_token_and_hmac(self):
        """Test building headers with both token and HMAC authentication."""
        with patch("athena_client.auth.get_settings") as mock_get_settings:
            mock_settings = Mock()
            mock_settings.ATHENA_TOKEN = "test-token"
            mock_settings.ATHENA_CLIENT_ID = "test-client"
            mock_settings.ATHENA_PRIVATE_KEY = (
                "-----BEGIN PRIVATE KEY-----\ntest-key\n-----END PRIVATE KEY-----"
            )
            mock_get_settings.return_value = mock_settings

            # Mock the imports that happen inside build_headers
            mock_datetime = Mock()
            mock_datetime.utcnow.return_value.strftime.return_value = (
                "2023-01-01T00:00:00Z"
            )

            mock_b64encode = Mock()
            mock_b64encode.return_value.decode.return_value = "test-signature"

            mock_serialization = Mock()
            mock_key = Mock()
            mock_serialization.load_pem_private_key.return_value = mock_key
            mock_key.sign.return_value = b"test-signature"

            mock_hashes = Mock()
            mock_hashes.SHA256.return_value = "sha256"

            with (
                patch("athena_client.auth.datetime", mock_datetime),
                patch("athena_client.auth.b64encode", mock_b64encode),
            ):
                build_headers(
                    "POST",
                    "https://api.example.com/test",
                    b"test-body",
                    serialization_module=mock_serialization,
                    hashes_module=mock_hashes,
                )
                # Verify the signing string includes the body
                expected_to_sign = "POST\nhttps://api.example.com/test\n\n2023-01-01T00:00:00Z\ntest-body"
                mock_key.sign.assert_called_once_with(
                    expected_to_sign.encode(), "sha256"
                )

    def test_build_headers_hmac_cryptography_import_error(self):
        """Test HMAC authentication when cryptography is not available."""
        with patch("athena_client.auth.get_settings") as mock_get_settings:
            mock_settings = Mock()
            mock_settings.ATHENA_TOKEN = None
            mock_settings.ATHENA_CLIENT_ID = None
            mock_settings.ATHENA_PRIVATE_KEY = "test-key"
            mock_get_settings.return_value = mock_settings

            # Patch sys.modules to simulate ImportError for cryptography and submodules
            import sys

            with patch.dict(
                sys.modules,
                {"cryptography": None, "cryptography.hazmat.primitives": None},
            ):
                with patch("athena_client.auth.logger") as mock_logger:
                    headers = build_headers("GET", "https://api.example.com/test", b"")
                    assert headers == {}
                    mock_logger.warning.assert_called_once()
                    mock_logger.warning.assert_called_once()

    def test_build_headers_hmac_exception(self):
        """Test HMAC authentication when an exception occurs."""
        with patch("athena_client.auth.get_settings") as mock_get_settings:
            mock_settings = Mock()
            mock_settings.ATHENA_TOKEN = None
            mock_settings.ATHENA_CLIENT_ID = None
            mock_settings.ATHENA_PRIVATE_KEY = "invalid-key"
            mock_get_settings.return_value = mock_settings

            # Mock the imports that happen inside build_headers
            mock_datetime = Mock()
            mock_datetime.utcnow.return_value.strftime.return_value = (
                "2023-01-01T00:00:00Z"
            )

            mock_serialization = Mock()
            mock_serialization.load_pem_private_key.side_effect = Exception(
                "Invalid key"
            )

            with patch("athena_client.auth.datetime", mock_datetime):
                with patch("athena_client.auth.logger") as mock_logger:
                    headers = build_headers(
                        "GET",
                        "https://api.example.com/test",
                        b"",
                        serialization_module=mock_serialization,
                        hashes_module=Mock(),
                    )

                    assert headers == {}
                    mock_logger.error.assert_called_once()

    def test_build_headers_different_methods(self):
        """Test building headers with different HTTP methods."""
        with patch("athena_client.auth.get_settings") as mock_get_settings:
            mock_settings = Mock()
            mock_settings.ATHENA_TOKEN = None
            mock_settings.ATHENA_CLIENT_ID = None
            mock_settings.ATHENA_PRIVATE_KEY = (
                "-----BEGIN PRIVATE KEY-----\ntest-key\n-----END PRIVATE KEY-----"
            )
            mock_get_settings.return_value = mock_settings

            # Mock the imports that happen inside build_headers
            mock_datetime = Mock()
            mock_datetime.utcnow.return_value.strftime.return_value = (
                "2023-01-01T00:00:00Z"
            )

            mock_b64encode = Mock()
            mock_b64encode.return_value.decode.return_value = "test-signature"

            mock_serialization = Mock()
            mock_key = Mock()
            mock_serialization.load_pem_private_key.return_value = mock_key
            mock_key.sign.return_value = b"test-signature"

            mock_hashes = Mock()
            mock_hashes.SHA256.return_value = "sha256"

            with (
                patch("athena_client.auth.datetime", mock_datetime),
                patch("athena_client.auth.b64encode", mock_b64encode),
            ):
                # Test different HTTP methods
                for method in ["GET", "POST", "PUT", "DELETE"]:
                    headers = build_headers(
                        method,
                        "https://api.example.com/test",
                        b"",
                        serialization_module=mock_serialization,
                        hashes_module=mock_hashes,
                    )
                    assert "X-Athena-Nonce" in headers
                    assert "X-Athena-Hmac" in headers
                    # Verify the signing string includes the method
                    expected_to_sign = f"{method}\nhttps://api.example.com/test\n\n2023-01-01T00:00:00Z\n"
                    mock_key.sign.assert_called_with(
                        expected_to_sign.encode(), "sha256"
                    )

    def test_build_headers_with_body(self):
        """Test building headers with request body."""
        with patch("athena_client.auth.get_settings") as mock_get_settings:
            mock_settings = Mock()
            mock_settings.ATHENA_TOKEN = None
            mock_settings.ATHENA_CLIENT_ID = None
            mock_settings.ATHENA_PRIVATE_KEY = (
                "-----BEGIN PRIVATE KEY-----\ntest-key\n-----END PRIVATE KEY-----"
            )
            mock_get_settings.return_value = mock_settings

            # Mock the imports that happen inside build_headers
            mock_datetime = Mock()
            mock_datetime.utcnow.return_value.strftime.return_value = (
                "2023-01-01T00:00:00Z"
            )

            mock_b64encode = Mock()
            mock_b64encode.return_value.decode.return_value = "test-signature"

            mock_serialization = Mock()
            mock_key = Mock()
            mock_serialization.load_pem_private_key.return_value = mock_key
            mock_key.sign.return_value = b"test-signature"

            mock_hashes = Mock()
            mock_hashes.SHA256.return_value = "sha256"

            with (
                patch("athena_client.auth.datetime", mock_datetime),
                patch("athena_client.auth.b64encode", mock_b64encode),
            ):
                build_headers(
                    "POST",
                    "https://api.example.com/test",
                    b"test-body",
                    serialization_module=mock_serialization,
                    hashes_module=mock_hashes,
                )
                # Verify the signing string includes the body
                expected_to_sign = "POST\nhttps://api.example.com/test\n\n2023-01-01T00:00:00Z\ntest-body"
                mock_key.sign.assert_called_once_with(
                    expected_to_sign.encode(), "sha256"
                )

    def test_build_headers_empty_body(self):
        """Test building headers with empty body."""
        with patch("athena_client.auth.get_settings") as mock_get_settings:
            mock_settings = Mock()
            mock_settings.ATHENA_TOKEN = None
            mock_settings.ATHENA_CLIENT_ID = None
            mock_settings.ATHENA_PRIVATE_KEY = (
                "-----BEGIN PRIVATE KEY-----\ntest-key\n-----END PRIVATE KEY-----"
            )
            mock_get_settings.return_value = mock_settings

            # Mock the imports that happen inside build_headers
            mock_datetime = Mock()
            mock_datetime.utcnow.return_value.strftime.return_value = (
                "2023-01-01T00:00:00Z"
            )

            mock_b64encode = Mock()
            mock_b64encode.return_value.decode.return_value = "test-signature"

            mock_serialization = Mock()
            mock_key = Mock()
            mock_serialization.load_pem_private_key.return_value = mock_key
            mock_key.sign.return_value = b"test-signature"

            mock_hashes = Mock()
            mock_hashes.SHA256.return_value = "sha256"

            with (
                patch("athena_client.auth.datetime", mock_datetime),
                patch("athena_client.auth.b64encode", mock_b64encode),
            ):
                build_headers(
                    "GET",
                    "https://api.example.com/test",
                    b"",
                    serialization_module=mock_serialization,
                    hashes_module=mock_hashes,
                )
                # Verify the signing string includes empty body
                expected_to_sign = (
                    "GET\nhttps://api.example.com/test\n\n2023-01-01T00:00:00Z\n"
                )
                mock_key.sign.assert_called_once_with(
                    expected_to_sign.encode(), "sha256"
                )



from typing import Dict, List, Protocol


class DatabaseConnector(Protocol):
    """Protocol for database connectors."""

    def validate_concepts(self, concept_ids: List[int]) -> List[int]:
        """Validate concept IDs against the local database."""
        ...

    def get_descendants(self, concept_ids: List[int]) -> List[int]:
        """Return descendant concept IDs for the given ancestors."""
        ...

    def get_standard_mapping(
        self, non_standard_concept_ids: List[int]
    ) -> Dict[int, int]:
        """Return mapping from non-standard to standard concept IDs."""
        ...
